/**
 * @file dist_grid.h
 *
 * @copyright Copyright  (C)  2019 Moritz Hanke <hanke@dkrz.de>
 *                                 Thomas Jahns <jahns@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Thomas Jahns <jahns@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DIST_GRID_H
#define DIST_GRID_H

#include <yaxt.h>

#include "geometry.h"
#include "proc_sphere_part.h"
#include "grid.h"

struct proc_sphere_part_node;
struct dist_grid_pair;
struct dist_grid;
struct interp_field;
typedef size_t const (* const const_size_t_2_pointer)[2];
typedef int const * const const_int_pointer;
typedef size_t const * const const_size_t_pointer;
typedef yac_int const * const const_yac_int_pointer;
typedef enum yac_edge_type const * const const_yac_edge_type_pointer;
typedef struct bounding_circle const * const const_bounding_circle_pointer;
typedef struct remote_point_infos const * const const_remote_point_infos_pointer;

struct const_basic_grid_data {
  const_coordinate_pointer vertex_coordinates;
  const_yac_int_pointer cell_ids;
  const_yac_int_pointer vertex_ids;
  const_yac_int_pointer edge_ids;
  const_int_pointer num_vertices_per_cell;
  const_size_t_pointer cell_to_vertex;
  const_size_t_pointer cell_to_vertex_offsets;
  const_size_t_pointer cell_to_edge;
  const_size_t_pointer cell_to_edge_offsets;
  const_size_t_2_pointer edge_to_vertex;
  const_bounding_circle_pointer cell_bnd_circles;
  const_yac_edge_type_pointer edge_type;
  const_remote_point_infos_pointer cell_owners;
  const_remote_point_infos_pointer vertex_owners;
  const_remote_point_infos_pointer edge_owners;
  const size_t num_cells;
  const size_t num_vertices;
  const size_t num_edges;
};

// information about a point on a remote process
struct remote_point_info {
  int rank;          // MPI rank
  uint64_t orig_pos; // local id on the remote process
};

struct remote_point_infos {
  int count;
  union {
    struct remote_point_info single; // valid if count == 1;
    struct remote_point_info * multi; // valid if count != -1;
  } data;
};

struct remote_point_info_reorder {
  struct remote_point_info data;
  size_t reorder_idx;
};

struct remote_point {
  yac_int global_id;
  struct remote_point_infos data;
};

struct remote_points {
  struct remote_point * data;
  size_t count;
  struct remote_point_info buffer[];
};

extern struct basic_grid_data basic_grid_data_empty;

struct dist_grid_pair * yac_dist_grid_pair_new(
  struct basic_grid_data * grid_a, int grid_id_a,
  struct basic_grid_data * grid_b, int grid_id_b,
  MPI_Comm comm);

void yac_dist_grid_pair_delete(struct dist_grid_pair * grid_pair);

MPI_Comm yac_dist_grid_pair_get_MPI_Comm(struct dist_grid_pair * grid_pair);

void yac_basic_grid_data_delete(struct basic_grid_data * grid);

struct dist_grid * yac_dist_grid_pair_get_dist_grid(
  struct dist_grid_pair * grid_pair, int grid_id);

void yac_dist_grid_pair_do_point_search(
  struct dist_grid_pair * grid_pair, int grid_id,
  coordinate_pointer search_coords, size_t count, size_t * src_cells,
  size_t num_fields, struct interp_field * fields);

void yac_dist_grid_pair_do_point_search_gc(
  struct dist_grid_pair * grid_pair, int grid_id,
  coordinate_pointer search_coords, size_t count, size_t * src_cells,
  size_t num_fields, struct interp_field * fields);

void yac_dist_grid_pair_do_nnn_search(
  struct dist_grid_pair * grid_pair, int grid_id,
  coordinate_pointer search_coords, size_t count, size_t * local_ids,
  size_t n, struct interp_field * field);

void yac_dist_grid_pair_do_bnd_circle_search(
  struct dist_grid_pair * grid_pair, int grid_id,
  const_bounding_circle_pointer bnd_circles, size_t count, size_t ** src_cells,
  size_t * num_src_per_tgt, struct interp_field * field);

void yac_dist_grid_pair_get_cell_neighbours(
  struct dist_grid_pair * grid_pair, int grid_id,
  size_t * cells, size_t count, size_t * neighbours,
  size_t num_fields, struct interp_field * fields);

struct remote_point_infos * yac_dist_grid_determine_remote_point_infos(
  struct dist_grid * grid, enum yac_location location, size_t count,
  yac_int * global_ids);

MPI_Comm yac_dist_grid_pair_get_global_comm(struct dist_grid_pair * grid_pair);

struct const_basic_grid_data *
yac_dist_grid_get_basic_grid_data(struct dist_grid * dist_grid);

size_t yac_dist_grid_get_local_count(
  struct dist_grid * dist_grid, enum yac_location location);

Xt_redist yac_dist_grid_get_redist_int(
  struct dist_grid * dist_grid, enum yac_location location);

/**
 * generates a redist that receives doubles to positions 0 to count-1, the
 * points to be received are specified by global ids
 * @param[in] grid       distributed grid
 * @param[in] location   data location
 * @param[in] global_ids points to be received
 * @param[in] count      size of global_ids
 * @returns redist
 */
Xt_redist yac_dist_grid_generate_redist_get_double(
  struct dist_grid * grid, enum yac_location location,
  yac_int * global_ids, size_t count);

/**
 * generates a redist that sends doubles from positions 0 to count-1 to
 * ranks and positions specified by point_infos
 * @param[in] grid        distributed grid
 * @param[in] location    data location
 * @param[in] point_infos send positions
 * @param[in] count       size of point_infos
 * @returns redist
 */
Xt_redist yac_dist_grid_generate_redist_put_double(
  struct dist_grid * grid, enum yac_location location,
  struct remote_point_infos * point_infos, size_t count);

void yac_dist_grid_get_local_unmasked_points(
  struct dist_grid * dist_grid, enum yac_location location, int * mask,
  size_t ** indices, size_t * count);

void yac_dist_grid_get_grid_cell(
  struct dist_grid * dist_grid, size_t cell_idx, struct grid_cell * cell);

void yac_dist_grid_determine_local_pos(
  struct dist_grid * dist_grid, enum yac_location location,
  yac_int * global_ids, size_t count, size_t * idx);

void yac_remote_point_infos_free(
  struct remote_point_infos * point_infos, size_t count);

MPI_Datatype yac_get_remote_point_info_mpi_datatype(MPI_Comm comm);

int yac_remote_point_get_pack_size(
  struct remote_point * point, MPI_Datatype point_info_dt, MPI_Comm comm);

void yac_remote_point_pack(
  struct remote_point * point, void * buffer, int buffer_size, int * position,
  MPI_Datatype point_info_dt, MPI_Comm comm);

void yac_remote_point_unpack(
  void * buffer, int buffer_size, int * position, struct remote_point * point,
  MPI_Datatype point_info_dt, MPI_Comm comm);

int yac_remote_points_get_pack_size(
  struct remote_points * points, MPI_Datatype point_info_dt, MPI_Comm comm);

void yac_remote_points_pack(
  struct remote_points * points, void * buffer, int buffer_size, int * position,
  MPI_Datatype point_info_dt, MPI_Comm comm);

void yac_remote_points_unpack(
  void * buffer, int buffer_size, int * position,
  struct remote_points ** points, MPI_Datatype point_info_dt, MPI_Comm comm);

struct remote_point * yac_dist_grid_get_remote_points(
  struct dist_grid * dist_grid, enum yac_location location,
  size_t * points, size_t count);

void yac_dist_grid_global_to_local(
  struct dist_grid * dist_grid, enum yac_location location,
  yac_int * global_ids, size_t count, size_t * local_ids,
  size_t num_fields, struct interp_field * fields);

void yac_dist_grid_get_cell_neighbours(
  struct dist_grid * grid, struct proc_sphere_part_node * proc_sphere_part,
  size_t * cells, size_t count, size_t * neighbours,
  size_t num_fields, struct interp_field * fields);

void yac_remote_point_infos_single_free(
  struct remote_point_infos * point_infos);

void yac_dist_grid_get_exchange_map(
  struct dist_grid * dist_grid,
    Xt_xmap * cell_xmap, Xt_xmap * vertex_xmap, Xt_xmap * edge_xmap);

size_t yac_dist_grid_get_total_num_points(
  struct dist_grid * grid, enum yac_location location);

void yac_dist_grid_pair_get_aux_grid_cells(
  struct dist_grid_pair * grid_pair, int grid_id,
  size_t * vertices, size_t count, size_t ** cells,
  int * num_cells_per_vertex, struct interp_field * field);

void yac_dist_grid_pair_get_aux_grid(
  struct dist_grid_pair * grid_pair, int grid_id,
  size_t * cells, size_t count,
  size_t ** vertex_to_cell, size_t ** vertex_to_cell_offsets,
  int ** num_cells_per_vertex, struct interp_field * field);

void yac_dist_grid_pair_relocate_point_pairs(
  struct dist_grid_pair * grid_pair, int a_is_ref,
  int grid_id_a, size_t ** points_a, enum yac_location location_a,
  struct interp_field * fields_a, size_t num_fields_a,
  int grid_id_b, size_t ** points_b, enum yac_location location_b,
  struct interp_field * fields_b, size_t num_fields_b, size_t * count);

#endif // DIST_GRID_H
