/**
 * @file interp_method_avg.c
 *
 * @copyright Copyright  (C)  2019 Moritz Hanke <hanke@dkrz.de>
 *                                 Thomas Jahns <jahns@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
// Get the definition of the 'restrict' keyword.
#include "config.h"
#endif

#include "string.h"
#include "grid_cell.h"
#include "interp_method_avg.h"

static size_t do_search_avg(struct interp_method * method,
                            struct interp_grid * interp_grid,
                            size_t * tgt_points, size_t count,
                            struct interp_weights * weights);
static void delete_avg(struct interp_method * method);

typedef int (*func_compute_weights)(
  double[3], size_t, const_coordinate_pointer, int*, double*);

static struct interp_method_vtable
  interp_method_avg_vtable = {
    .do_search = do_search_avg,
    .delete = delete_avg};

struct interp_method_avg {

  struct interp_method_vtable * vtable;
  func_compute_weights compute_weights;
};

static int check_src_field_cell(
  double * point, size_t field_cell, size_t cell_size,
  size_t const * vertex_to_cell, size_t const * vertex_to_cell_offsets,
  const_coordinate_pointer field_coordinates, struct grid_cell * cell_buffer) {

  if (cell_size == 0) return 0;

  cell_buffer->num_corners = cell_size;

  coordinate_pointer coordinates_xyz;

  if (cell_size > cell_buffer->array_size) {
    cell_buffer->coordinates_xyz =
      ((coordinates_xyz =
          xrealloc(
            cell_buffer->coordinates_xyz,
            cell_size * sizeof(*coordinates_xyz))));
    cell_buffer->edge_type =
      xrealloc(
        cell_buffer->edge_type, cell_size * sizeof(cell_buffer->edge_type));
    for (size_t i = 0; i < cell_size; ++i)
      cell_buffer->edge_type[i] = GREAT_CIRCLE;
  } else {
    coordinates_xyz = cell_buffer->coordinates_xyz;
  }

  size_t const * field_cell_vertices =
    vertex_to_cell + vertex_to_cell_offsets[field_cell];
  for (size_t i = 0; i < cell_size; ++i)
    memcpy(coordinates_xyz[i], field_coordinates[field_cell_vertices[i]],
           sizeof(*coordinates_xyz));

  return yac_point_in_cell(point, *cell_buffer);
}

static size_t do_search_avg (struct interp_method * method,
                             struct interp_grid * interp_grid,
                             size_t * tgt_points, size_t count,
                             struct interp_weights * weights) {

  struct interp_method_avg * method_avg = (struct interp_method_avg *)method;
  func_compute_weights compute_weights = method_avg->compute_weights;

  if (yac_interp_grid_get_num_src_fields(interp_grid) != 1)
    yac_internal_abort_message(
      "ERROR(do_search_avg): invalid number of source fields",
      __FILE__, __LINE__);

  enum yac_location src_field_location =
    yac_interp_grid_get_src_field_location(interp_grid, 0);

  if ((src_field_location != CORNER) && (src_field_location != CELL))
    yac_internal_abort_message(
      "ERROR(do_search_avg): unsupported source field location type",
      __FILE__, __LINE__);

  // get coordinates of target points
  coordinate_pointer tgt_coords = xmalloc(count * sizeof(*tgt_coords));
  yac_interp_grid_get_tgt_coordinates(
    interp_grid, tgt_points, count, tgt_coords);

  size_t * size_t_buffer = xmalloc(2 * count * sizeof(*size_t_buffer));
  size_t * src_field_cells = size_t_buffer;
  size_t * reorder_idx = size_t_buffer + count;

  // get matching source cells for all target points
  yac_interp_grid_do_points_search(
    interp_grid, tgt_coords, count, src_field_cells);

  size_t const * src_field_cell_to_vertex;
  size_t const * src_field_cell_to_vertex_offsets;
  int const * src_field_num_vertices_per_cell;

  // if the source field is location at cell points
  if (src_field_location == CELL) {

    // generate auxiliary grid for all search result cells
    yac_interp_grid_get_aux_grid_src(
      interp_grid, src_field_cells, count,
      (size_t**)&src_field_cell_to_vertex,
      (size_t**)&src_field_cell_to_vertex_offsets,
      (int**)&src_field_num_vertices_per_cell);

    struct const_basic_grid_data * src_grid_data =
      yac_interp_grid_get_basic_grid_data_src(interp_grid);
    const_coordinate_pointer src_field_coordinates =
      yac_interp_grid_get_src_field_coords(interp_grid, 0);

    struct grid_cell cell_buffer;
    yac_init_grid_cell(&cell_buffer);

    // for all target point
    for (size_t i = 0; i < count; ++i) {

      size_t curr_src_cell = src_field_cells[i];
      if (curr_src_cell == SIZE_MAX) continue;

      double * curr_tgt_coord = tgt_coords[i];
      size_t const * curr_vertices =
        src_grid_data->cell_to_vertex +
        src_grid_data->cell_to_vertex_offsets[curr_src_cell];
      size_t curr_num_vertices =
        src_grid_data->num_vertices_per_cell[curr_src_cell];

      size_t result_src_field_cell = SIZE_MAX;

      // for all auxiliary of the current source result cell
      for (size_t j = 0; j < curr_num_vertices; ++j) {

        size_t src_field_cell = curr_vertices[j];
        size_t src_field_cell_size =
          (size_t)src_field_num_vertices_per_cell[src_field_cell];

        // check whether the target point is in the current source field cell
        if (check_src_field_cell(
              curr_tgt_coord, src_field_cell, src_field_cell_size,
              src_field_cell_to_vertex, src_field_cell_to_vertex_offsets,
              src_field_coordinates, &cell_buffer)) {
          result_src_field_cell = src_field_cell;
          break;
        }
      }

      src_field_cells[i] = result_src_field_cell;
    }
    yac_free_grid_cell(&cell_buffer);
  } else {

    struct const_basic_grid_data * grid_data =
      yac_interp_grid_get_basic_grid_data_src(interp_grid);
    src_field_cell_to_vertex = grid_data->cell_to_vertex;
    src_field_cell_to_vertex_offsets = grid_data->cell_to_vertex_offsets;
    src_field_num_vertices_per_cell = grid_data->num_vertices_per_cell;
  }
  const_coordinate_pointer src_field_coordinates =
    yac_interp_grid_get_src_field_coords(interp_grid, 0);

  // sort target points, for which we found a source cell, to the beginning of
  // the array
  for (size_t i = 0; i < count; ++i) reorder_idx[i] = i;
  yac_quicksort_index_size_t_size_t(src_field_cells, count, reorder_idx);

  size_t result_count = 0;
  for (result_count = 0; result_count < count; ++result_count)
    if (src_field_cells[result_count] == SIZE_MAX) break;

  size_t total_num_weights = 0;
  size_t max_num_vertices_per_cell = 0;
  size_t * num_weights_per_tgt =
    xmalloc(result_count * sizeof(*num_weights_per_tgt));

  // get the number of vertices per target cell and the total number of vertices
  // required for the interpolation (may contain duplicated vertices)
  for (size_t i = 0; i < result_count; ++i) {
    size_t curr_num_vertices =
      (size_t)(src_field_num_vertices_per_cell[src_field_cells[i]]);
    num_weights_per_tgt[i] = curr_num_vertices;
    total_num_weights += curr_num_vertices;
    if (curr_num_vertices > max_num_vertices_per_cell)
      max_num_vertices_per_cell = curr_num_vertices;
  }
  if (src_field_location == CELL)
    free((void*)src_field_num_vertices_per_cell);

  const_int_pointer src_field_mask =
    yac_interp_grid_get_src_field_mask(interp_grid, 0);

  coordinate_pointer src_coord_buffer =
    xmalloc(max_num_vertices_per_cell * sizeof(*src_coord_buffer));
  int * mask_buffer =
    (src_field_mask != NULL)?
      xmalloc(max_num_vertices_per_cell * sizeof(*mask_buffer)):NULL;
  double * w = xmalloc(total_num_weights * sizeof(*w));
  size_t * src_points = xmalloc(total_num_weights * sizeof(*src_points));

  // for each target point, extract relevant source data and compute the weights
  // based on that
  total_num_weights = 0;
  for (size_t i = 0; i < result_count;) {

    size_t curr_num_vertices = num_weights_per_tgt[i];

    const_size_t_pointer curr_cell_to_vertex =
      src_field_cell_to_vertex +
      src_field_cell_to_vertex_offsets[src_field_cells[i]];
    double * curr_weights = w + total_num_weights;

    // get the mask and the source coordinates of the source vertices required
    // for the current target point
    for (size_t j = 0; j < curr_num_vertices; ++j, ++total_num_weights) {

      size_t curr_vertex_idx = curr_cell_to_vertex[j];
      src_points[total_num_weights] = curr_vertex_idx;
      for (size_t k = 0; k < 3; ++k)
        src_coord_buffer[j][k] = src_field_coordinates[curr_vertex_idx][k];
      if (src_field_mask != NULL)
        mask_buffer[j] = src_field_mask[curr_vertex_idx];
    }

    // compute the weights
    if (compute_weights(tgt_coords[reorder_idx[i]], curr_num_vertices,
                        (const_coordinate_pointer)src_coord_buffer, mask_buffer,
                        curr_weights)) {

      // if weight computation was successful
      num_weights_per_tgt[i] = curr_num_vertices;
      ++i;
    } else {

      // if no weights could be computed
      total_num_weights -= curr_num_vertices;
      result_count--;
      src_field_cells[i] = src_field_cells[result_count];
      size_t temp_reorder_idx = reorder_idx[i];
      reorder_idx[i] = reorder_idx[result_count];
      reorder_idx[result_count] = temp_reorder_idx;
    }
  }
  if (src_field_location == CELL) {
    free((void*)src_field_cell_to_vertex);
    free((void*)src_field_cell_to_vertex_offsets);
  }

  // move the non-interpolated target points to the end
  for (size_t i = 0; i < count; ++i) src_field_cells[reorder_idx[i]] = i;
  yac_quicksort_index_size_t_size_t(src_field_cells, count, tgt_points);

  free(tgt_coords);
  free(mask_buffer);
  free(src_coord_buffer);
  free(size_t_buffer);

  struct remote_points tgts = {
    .data =
      yac_interp_grid_get_tgt_remote_points(
        interp_grid, tgt_points, result_count),
    .count = result_count};
  struct remote_point * srcs =
    yac_interp_grid_get_src_remote_points(
      interp_grid, 0, src_points, total_num_weights);

  // store weights
  yac_interp_weights_add_wsum(
    weights, &tgts, num_weights_per_tgt, srcs, w);

  free(tgts.data);
  free(srcs);
  free(w);
  free(num_weights_per_tgt);
  free(src_points);

  return result_count;
}

static int compute_weights_avg_yes(
  double tgt_coords[3], size_t num_vertices,
  const_coordinate_pointer src_coords, int * src_mask, double * weights) {

  size_t num_unmasked_points = 0;

  // if we have a source mask, check whether any points is masked
  if (src_mask != NULL)
    for (size_t i = 0; i < num_vertices; ++i)
      if (src_mask[i]) num_unmasked_points++;

  if (num_unmasked_points > 0) {

    double weight = 1.0 / (double)num_unmasked_points;
    for (size_t i = 0; i < num_vertices; ++i)
      weights[i] = (src_mask[i])?weight:0.0;
  }

  return num_unmasked_points > 0;
}

static int compute_weights_avg_no(
  double tgt_coords[3], size_t num_vertices,
  const_coordinate_pointer src_coords, int * src_mask, double * weights) {

  // if we have a source mask, check whether any points is masked
  if (src_mask != NULL)
    for (size_t i = 0; i < num_vertices; ++i)
      if (!src_mask[i]) return 0;

  double weight = 1.0 / (double)num_vertices;
  for (size_t i = 0; i < num_vertices; ++i) weights[i] = weight;

  return 1;
}

static int compute_weights_dist_yes(
  double tgt_coords[3], size_t num_vertices,
  const_coordinate_pointer src_coords, int * src_mask, double * weights) {

  // if there is a source mask, check if there are unmasked vertices
  if (src_mask != NULL) {
    int has_unmasked = 0;
    for (size_t i = 0; i < num_vertices; ++i) has_unmasked |= src_mask[i];
    if (!has_unmasked) return 0;
  }

  if (src_mask != NULL) {
    for (size_t i = 0; i < num_vertices; ++i) {

      if (src_mask[i]) {

        double distance =
          get_vector_angle(tgt_coords, (double*)(src_coords[i]));

        // if the target and source point are nearly identical
        if (distance < yac_angle_tol) {
          for (size_t j = 0; j < num_vertices; ++j) weights[j] = 0.0;
          weights[i] = 1.0;
          return 1;
        }

        weights[i] = 1.0 / distance;
      } else {
        weights[i] = 0.0;
      }
    }
  } else {
    for (size_t i = 0; i < num_vertices; ++i) {

      double distance =
        get_vector_angle(tgt_coords, (double*)(src_coords[i]));

      // if the target and source point are nearly identical
      if (distance < yac_angle_tol) {
        for (size_t j = 0; j < num_vertices; ++j) weights[j] = 0.0;
        weights[i] = 1.0;
        return 1;
      }

      weights[i] = 1.0 / distance;
    }
  }

  // compute scaling factor for the weights
  double inv_distance_sum = 0.0;
  for (size_t i = 0; i < num_vertices; ++i)
    inv_distance_sum += weights[i];
  double scale = 1.0 / inv_distance_sum;

  for (size_t i = 0; i < num_vertices; ++i) weights[i] *= scale;

  return 1;
}

static int compute_weights_dist_no(
  double tgt_coords[3], size_t num_vertices,
  const_coordinate_pointer src_coords, int * src_mask, double * weights) {

  // if there is a source mask, check if there are masked vertices
  if (src_mask != NULL)
    for (size_t i = 0; i < num_vertices; ++i) if(!src_mask[i]) return 0;

  for (size_t i = 0; i < num_vertices; ++i) {

    double distance =
      get_vector_angle(tgt_coords, (double*)(src_coords[i]));

    // if the target and source point are nearly identical
    if (distance < yac_angle_tol) {
      for (size_t j = 0; j < num_vertices; ++j) weights[j] = 0.0;
      weights[i] = 1.0;
      return 1;
    }

    weights[i] = 1.0 / distance;
  }

  // compute scaling factor for the weights
  double inv_distance_sum = 0.0;
  for (size_t i = 0; i < num_vertices; ++i)
    inv_distance_sum += weights[i];
  double scale = 1.0 / inv_distance_sum;

  for (size_t i = 0; i < num_vertices; ++i) weights[i] *= scale;

  return 1;
}

static func_compute_weights select_compute_weight_routine(
  enum yac_interp_avg_weight_type weight_type, int partial_coverage) {

  switch(weight_type) {
    case(AVG):
      if (partial_coverage) return compute_weights_avg_yes;
      else                  return compute_weights_avg_no;
    case(DIST):
      if (partial_coverage) return compute_weights_dist_yes;
      else                  return compute_weights_dist_no;
    default:
      yac_internal_abort_message(
        "ERROR(select_compute_weight_routine): invalid weight type",
        __FILE__, __LINE__);
      return NULL;
  };
}

struct interp_method * yac_interp_method_avg_new(
  enum yac_interp_avg_weight_type weight_type,
  int partial_coverage) {

  struct interp_method_avg * method = xmalloc(1 * sizeof(*method));

  method->vtable = &interp_method_avg_vtable;
  method->compute_weights =
    select_compute_weight_routine(weight_type, partial_coverage);

  return (struct interp_method*)method;
}

static void delete_avg(struct interp_method * method) {
  free(method);
}
