/**
 * @file test_couple_config_parallel1.c
 *
 * @copyright Copyright  (C)  2015 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "tests.h"
#include "test_common.h"
#include "xml/config_xml.h"
#include "couple_config.h"
#include "utils.h"
#include "yac_interface.h"
#include "yac_mpi.h"
#include "dist_grid_utils.h"

int main (void) {

  yac_mpi_call(MPI_Init(NULL, NULL), MPI_COMM_WORLD);
  yac_mpi_call(
    MPI_Comm_dup(MPI_COMM_WORLD, &yac_mpi_comm_world), MPI_COMM_WORLD);
  xt_initialize(yac_mpi_comm_world);
  int rank, size;
  yac_mpi_call(
    MPI_Comm_rank(yac_mpi_comm_world, &rank), yac_mpi_comm_world);
  yac_mpi_call(
    MPI_Comm_size(yac_mpi_comm_world, &size), yac_mpi_comm_world);

  if (size != 3) {

    PUT_ERR("ERROR: wrong number of processes\n");
    return TEST_EXIT_CODE;
  }

  { // tests with couple_config_test_1_1.xml
    yac_xml_init_config_xml("couple_config_test_1_1.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // no process defines a components
      yac_setup_couple_config(0, NULL, 0, NULL);
    }

    { // first two process define component comp_1 and last defines comp_2

      struct component * components[1];
      char * component_names[2] = {"comp_1", "comp_2"};
      unsigned component_id;

      components[0] = yac_component_new(component_names[rank >> 1]);
      component_id = yac_pointer_to_unique_id(components[0]);

      size_t num_vertices[2] = {2,2};
      int cyclic[2] = {0,0};
      double coordinates_x[3][2] = {{0,1},{1,2},{0.5,1.5}};
      double coordinates_y[3][2] = {{0,1},{0,1},{0,1}};

      char * grid_name[2] = {"grid1", "grid2"};
      struct basic_grid grid;
      grid.name = grid_name[rank>>1];
      grid.data =
        yac_generate_basic_grid_data_reg_2d_deg(
          num_vertices, cyclic, coordinates_x[rank], coordinates_y[rank]);

      yac_int global_cell_ids [3][1] = {{0},{1},{0}};
      yac_int global_corner_ids[3][4] = {{0,1,3,4},{1,2,4,5},{0,1,2,3}};
      int cell_core_mask[1] = {1};
      int corner_core_mask[3][4] = {{1,1,1,0},{0,1,1,1},{1,1,1,1}};

      grid.data.cell_ids = TO_POINTER(global_cell_ids[rank]);
      grid.data.vertex_ids = TO_POINTER(global_corner_ids[rank]);
      grid.data.core_cell_mask = TO_POINTER(cell_core_mask);
      grid.data.core_vertex_mask = TO_POINTER(corner_core_mask[rank]);

      struct interp_field interp_fields[1];
      interp_fields[0].location = CORNER;
      interp_fields[0].coordinates = NULL;
      interp_fields[0].mask = NULL;

      struct coupling_field * coupling_fields[1];
      char * field_name = "transient_1";

      coupling_fields[0] =
        yac_coupling_field_new(
          field_name, component_id, &grid, interp_fields, 1);

      yac_setup_couple_config(1, components, 1, coupling_fields);

      yac_coupling_field_delete(coupling_fields[0]);
      yac_basic_grid_data_free(grid.data);
      yac_component_delete(components[0]);
    }

    yac_components_reset();
    yac_xml_destroy_config_xml();
  }

  { // tests with couple_config_test_1_2.xml
    yac_xml_init_config_xml("couple_config_test_1_2.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // each process has its own component and data exchange occurs in a
      // round robin fashion

      struct component * components[1];
      char * component_names[3] = {"comp_1", "comp_2", "comp_3"};
      unsigned component_id;

      components[0] = yac_component_new(component_names[rank]);
      component_id = yac_pointer_to_unique_id(components[0]);

      size_t num_vertices[2] = {2,2};
      int cyclic[2] = {0,0};
      double coordinates_x[2] = {0,1};
      double coordinates_y[2] = {0,1};

      char * grid_name[3] = {"grid1", "grid2", "grid3"};
      struct basic_grid grid;
      grid.name = grid_name[rank];
      grid.data =
        yac_generate_basic_grid_data_reg_2d_deg(
          num_vertices, cyclic, coordinates_x, coordinates_y);

      yac_int global_cell_ids [1] = {0};
      yac_int global_corner_ids[4] = {0,1,2,3};
      int cell_core_mask[1] = {1};
      int corner_core_mask[4] = {1,1,1,1};

      grid.data.cell_ids = TO_POINTER(global_cell_ids);
      grid.data.vertex_ids = TO_POINTER(global_corner_ids);
      grid.data.core_cell_mask = TO_POINTER(cell_core_mask);
      grid.data.core_vertex_mask = TO_POINTER(corner_core_mask);

      struct interp_field interp_fields[1];
      interp_fields[0].location = CORNER;
      interp_fields[0].coordinates = NULL;
      interp_fields[0].mask = NULL;

      struct coupling_field * coupling_fields[2];
      char * field_name[3] = {"transient_1", "transient_2", "transient_3"};

      coupling_fields[0] =
        yac_coupling_field_new(
          field_name[rank], component_id, &grid, interp_fields, 1);
      coupling_fields[1] =
        yac_coupling_field_new(
          field_name[(rank + 2) % 3], component_id, &grid, interp_fields, 1);

      yac_setup_couple_config(1, components, 2, coupling_fields);

      yac_coupling_field_delete(coupling_fields[0]);
      yac_coupling_field_delete(coupling_fields[1]);
      yac_basic_grid_data_free(grid.data);
      yac_component_delete(components[0]);
    }

    yac_components_reset();
    yac_xml_destroy_config_xml();
  }

  { // tests with couple_config_test_1_3.xml
    yac_xml_init_config_xml("couple_config_test_1_3.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // each process has its own component and data exchange occurs in a
      // round robin fashion

      struct component * components[1];
      char * component_names[3] = {"comp_1", "comp_2", "comp_3"};
      unsigned component_id;

      components[0] = yac_component_new(component_names[rank]);
      component_id = yac_pointer_to_unique_id(components[0]);

      size_t num_vertices[2] = {2,2};
      int cyclic[2] = {0,0};
      double coordinates_x[2] = {0,1};
      double coordinates_y[2] = {0,1};

      char * grid_name[3] = {"grid1", "grid2", "grid3"};
      struct basic_grid grid;
      grid.name = grid_name[rank];
      grid.data =
        yac_generate_basic_grid_data_reg_2d_deg(
          num_vertices, cyclic, coordinates_x, coordinates_y);

      yac_int global_cell_ids [1] = {0};
      yac_int global_corner_ids[4] = {0,1,2,3};
      int cell_core_mask[1] = {1};
      int corner_core_mask[4] = {1,1,1,1};

      grid.data.cell_ids = TO_POINTER(global_cell_ids);
      grid.data.vertex_ids = TO_POINTER(global_corner_ids);
      grid.data.core_cell_mask = TO_POINTER(cell_core_mask);
      grid.data.core_vertex_mask = TO_POINTER(corner_core_mask);

      struct interp_field interp_fields[1];
      interp_fields[0].location = CORNER;
      interp_fields[0].coordinates = NULL;
      interp_fields[0].mask = NULL;

      struct coupling_field * coupling_fields[4];
      char * field_name[6] = {"transient_1", "transient_2", "transient_3",
                              "transient_4", "transient_5", "transient_6"};

      coupling_fields[0] =
        yac_coupling_field_new(
          field_name[rank], component_id, &grid, interp_fields, 1);
      coupling_fields[1] =
        yac_coupling_field_new(
          field_name[(rank + 2) % 3], component_id, &grid, interp_fields, 1);
      coupling_fields[2] =
        yac_coupling_field_new(
          field_name[rank + 3], component_id, &grid, interp_fields, 1);
      coupling_fields[3] =
        yac_coupling_field_new(
          field_name[((rank + 2) % 3) + 3], component_id, &grid, interp_fields, 1);

      yac_setup_couple_config(1, components, 4, coupling_fields);

      yac_coupling_field_delete(coupling_fields[0]);
      yac_coupling_field_delete(coupling_fields[1]);
      yac_coupling_field_delete(coupling_fields[2]);
      yac_coupling_field_delete(coupling_fields[3]);
      yac_basic_grid_data_free(grid.data);
      yac_component_delete(components[0]);
    }

    yac_components_reset();
    yac_xml_destroy_config_xml();
  }

  { // tests with couple_config_test_1_4.xml
    yac_xml_init_config_xml("couple_config_test_1_4.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // each process has its own component
      // process one has one field that is to be sent to both other
      // processes in the same put

      struct component * components[1];
      char * component_names[3] = {"comp_1", "comp_2", "comp_3"};
      unsigned component_id;

      components[0] = yac_component_new(component_names[rank]);
      component_id = yac_pointer_to_unique_id(components[0]);

      size_t num_vertices[2] = {2,2};
      int cyclic[2] = {0,0};
      double coordinates_x[2] = {0,1};
      double coordinates_y[2] = {0,1};

      char * grid_name[3] = {"grid1", "grid2", "grid3"};
      struct basic_grid grid;
      grid.name = grid_name[rank];
      grid.data =
        yac_generate_basic_grid_data_reg_2d_deg(
          num_vertices, cyclic, coordinates_x, coordinates_y);

      yac_int global_cell_ids [1] = {0};
      yac_int global_corner_ids[4] = {0,1,2,3};
      int cell_core_mask[1] = {1};
      int corner_core_mask[4] = {1,1,1,1};

      grid.data.cell_ids = TO_POINTER(global_cell_ids);
      grid.data.vertex_ids = TO_POINTER(global_corner_ids);
      grid.data.core_cell_mask = TO_POINTER(cell_core_mask);
      grid.data.core_vertex_mask = TO_POINTER(corner_core_mask);

      struct interp_field interp_fields[1];
      interp_fields[0].location = CORNER;
      interp_fields[0].coordinates = NULL;
      interp_fields[0].mask = NULL;

      struct coupling_field * coupling_fields[1];
      char * field_name = "transient_1";

      coupling_fields[0] =
        yac_coupling_field_new(
          field_name, component_id, &grid, interp_fields, 1);

      yac_setup_couple_config(1, components, 1, coupling_fields);

      yac_coupling_field_delete(coupling_fields[0]);
      yac_basic_grid_data_free(grid.data);
      yac_component_delete(components[0]);
    }

    yac_components_reset();
    yac_xml_destroy_config_xml();
  }

  { // tests with couple_config_test_1_5.xml
    yac_xml_init_config_xml("couple_config_test_1_5.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // each process has its own component
      // process one has one field that is to be sent to both other
      // processes in the same put

      struct component * components[1];
      char * component_names[3] = {"comp_1", "comp_2", "comp_3"};
      unsigned component_id;

      components[0] = yac_component_new(component_names[rank]);
      component_id = yac_pointer_to_unique_id(components[0]);

      size_t num_vertices[2] = {5,5};
      int cyclic[2] = {0,0};
      double coordinates_x[5] = {0,1,2,3,4};
      double coordinates_y[5] = {0,1,2,3,4};

      char * grid_name[3] = {"grid1", "grid2", "grid3"};
      struct basic_grid grid;
      grid.name = grid_name[rank];
      grid.data =
        yac_generate_basic_grid_data_reg_2d_deg(
          num_vertices, cyclic, coordinates_x, coordinates_y);

      struct interp_field interp_fields[1];
      interp_fields[0].location = CORNER;
      interp_fields[0].coordinates = NULL;
      interp_fields[0].mask = NULL;

      struct coupling_field * coupling_fields[1];
      char * field_name = "transient_1";

      coupling_fields[0] =
        yac_coupling_field_new(
          field_name, component_id, &grid, interp_fields, 1);

      yac_setup_couple_config(1, components, 1, coupling_fields);

      yac_coupling_field_delete(coupling_fields[0]);
      yac_basic_grid_data_free(grid.data);
      yac_component_delete(components[0]);
    }

    yac_components_reset();
    yac_xml_destroy_config_xml();
  }

#ifdef YAC_NETCDF_ENABLED
  { // tests with couple_config_test_1_6.xml
    yac_xml_init_config_xml("couple_config_test_1_6.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // first two process define component comp_1 and last defines comp_2

      struct component * components[1];
      char * component_names[2] = {"comp_1", "comp_2"};
      unsigned component_id;

      components[0] = yac_component_new(component_names[rank >> 1]);
      component_id = yac_pointer_to_unique_id(components[0]);

      size_t num_vertices[2] = {2,2};
      int cyclic[2] = {0,0};
      double coordinates_x[3][2] = {{0,1},{1,2},{0.5,1.5}};
      double coordinates_y[3][2] = {{0,1},{0,1},{0,1}};

      char * grid_name[2] = {"grid1", "grid2"};
      struct basic_grid grid;
      grid.name = grid_name[rank>>1];
      grid.data =
        yac_generate_basic_grid_data_reg_2d_deg(
          num_vertices, cyclic, coordinates_x[rank], coordinates_y[rank]);

      yac_int global_cell_ids [3][1] = {{0},{1},{0}};
      yac_int global_corner_ids[3][4] = {{0,1,3,4},{1,2,4,5},{0,1,2,3}};
      int cell_core_mask[1] = {1};
      int corner_core_mask[3][4] = {{1,1,1,0},{0,1,1,1},{1,1,1,1}};

      grid.data.cell_ids = TO_POINTER(global_cell_ids[rank]);
      grid.data.vertex_ids = TO_POINTER(global_corner_ids[rank]);
      grid.data.core_cell_mask = TO_POINTER(cell_core_mask);
      grid.data.core_vertex_mask = TO_POINTER(corner_core_mask[rank]);

      struct interp_field interp_fields[1];
      interp_fields[0].location = CORNER;
      interp_fields[0].coordinates = NULL;
      interp_fields[0].mask = NULL;

      struct coupling_field * coupling_fields[1];
      char * field_name = "transient_1";

      char * weight_file_name = "weight_file.nc";

      // delete weight file if it extists
      if (rank == 0) unlink(weight_file_name);

      coupling_fields[0] =
        yac_coupling_field_new(
          field_name, component_id, &grid, interp_fields, 1);

      yac_setup_couple_config(1, components, 1, coupling_fields);

      // ensure that the weight file has been written
      MPI_Barrier(MPI_COMM_WORLD);

      // check whether weight file exists
      if (rank == 0) {
        if (access(weight_file_name, F_OK ) == -1)
          PUT_ERR("weight file is missing\n");
        // delete weight file if it extists
        unlink(weight_file_name);
      }

      yac_coupling_field_delete(coupling_fields[0]);
      yac_basic_grid_data_free(grid.data);
      yac_component_delete(components[0]);
    }

    yac_components_reset();
    yac_xml_destroy_config_xml();
  }
#endif

#ifdef YAC_NETCDF_ENABLED
  { // tests with couple_config_test_1_7.xml
    yac_xml_init_config_xml("couple_config_test_1_7.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // first two process define component comp_1 and last defines comp_2

      struct component * components[1];
      char * component_names[2] = {"comp_1", "comp_2"};
      unsigned component_id;

      components[0] = yac_component_new(component_names[rank >> 1]);
      component_id = yac_pointer_to_unique_id(components[0]);

      size_t num_vertices[2] = {2,2};
      int cyclic[2] = {0,0};
      double coordinates_x[3][2] = {{0,1},{1,2},{0.5,1.5}};
      double coordinates_y[3][2] = {{0,1},{0,1},{0,1}};

      char * grid_name[2] = {"grid1", "grid2"};
      struct basic_grid grid;
      grid.name = grid_name[rank>>1];
      grid.data =
        yac_generate_basic_grid_data_reg_2d_deg(
          num_vertices, cyclic, coordinates_x[rank], coordinates_y[rank]);

      yac_int global_cell_ids [3][1] = {{0},{1},{0}};
      yac_int global_corner_ids[3][4] = {{0,1,3,4},{1,2,4,5},{0,1,2,3}};
      int cell_core_mask[1] = {1};
      int corner_core_mask[3][4] = {{1,1,1,0},{0,1,1,1},{1,1,1,1}};

      grid.data.cell_ids = TO_POINTER(global_cell_ids[rank]);
      grid.data.vertex_ids = TO_POINTER(global_corner_ids[rank]);
      grid.data.core_cell_mask = TO_POINTER(cell_core_mask);
      grid.data.core_vertex_mask = TO_POINTER(corner_core_mask[rank]);

      struct interp_field interp_fields[1];
      interp_fields[0].location = CORNER;
      interp_fields[0].coordinates = NULL;
      interp_fields[0].mask = NULL;

      struct coupling_field * coupling_fields[3];
      char * field_name[3] = {"transient_1", "transient_2", "transient_3"};

      char * weight_file_name[3] =
        {NULL, "weight_file_1.nc", "weight_file_2.nc"};

      // delete weight file if it extists
      for (unsigned i = 0; i < 3; ++i) {

        if ((rank == 0) && (weight_file_name[i] != NULL))
          unlink(weight_file_name[i]);

        coupling_fields[i] =
          yac_coupling_field_new(
            field_name[i], component_id, &grid, interp_fields, 1);
      }

      yac_setup_couple_config(1, components, 3, coupling_fields);

      // ensure that the weight file has been written
      MPI_Barrier(MPI_COMM_WORLD);

      // check whether weight file exists
      if (rank == 0) {
        for (unsigned i = 0; i < 3; ++i) {
          if (weight_file_name[i] == NULL) continue;
          if (access(weight_file_name[i], F_OK ) == -1)
            PUT_ERR("weight file is missing\n");
          // delete weight file if it extists
          unlink(weight_file_name[i]);
        }
      }

      for (unsigned i = 0; i < 3; ++i)
        yac_coupling_field_delete(coupling_fields[i]);
      yac_basic_grid_data_free(grid.data);
      yac_component_delete(components[0]);
    }

    yac_components_reset();
    yac_xml_destroy_config_xml();
  }
#endif


  { // tests with couple_config_test_1_8.xml
    yac_xml_init_config_xml("couple_config_test_1_8.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // all processes define comp_1, but the xml lists comp_1 and comp_2

      struct component * component = yac_component_new("comp_1");
      unsigned component_id = yac_pointer_to_unique_id(component);

      size_t num_vertices[2] = {2,2};
      int cyclic[2] = {0,0};
      double coordinates_x[3][2] = {{0,1},{1,2},{2,3}};
      double coordinates_y[3][2] = {{0,1},{0,1},{0,1}};

      char * grid_name[1] = {"grid1"};
      struct basic_grid grid;
      grid.name = grid_name[1];
      grid.data =
        yac_generate_basic_grid_data_reg_2d_deg(
          num_vertices, cyclic, coordinates_x[rank], coordinates_y[rank]);

      yac_int global_cell_ids [3][1] = {{0},{1},{2}};
      yac_int global_corner_ids[3][4] = {{0,1,4,5},{1,2,5,6},{2,3,6,7}};
      int cell_core_mask[1] = {1};
      int corner_core_mask[3][4] = {{1,1,1,0},{0,1,1,0},{0,1,1,1}};

      grid.data.cell_ids = TO_POINTER(global_cell_ids[rank]);
      grid.data.vertex_ids = TO_POINTER(global_corner_ids[rank]);
      grid.data.core_cell_mask = TO_POINTER(cell_core_mask);
      grid.data.core_vertex_mask = TO_POINTER(corner_core_mask[rank]);

      struct interp_field interp_fields[1];
      interp_fields[0].location = CORNER;
      interp_fields[0].coordinates = NULL;
      interp_fields[0].mask = NULL;

      struct coupling_field * coupling_fields[3];
      char * field_name[3] = {"transient_1", "transient_2", "transient_3"};

      // delete weight file if it extists
      for (unsigned i = 0; i < 3; ++i) {

        coupling_fields[i] =
          yac_coupling_field_new(
            field_name[i], component_id, &grid, interp_fields, 1);
      }

      yac_setup_couple_config(1, &component, 3, coupling_fields);

      yac_coupling_field_delete(coupling_fields[0]);
      yac_coupling_field_delete(coupling_fields[1]);
      yac_coupling_field_delete(coupling_fields[2]);
      yac_basic_grid_data_free(grid.data);
      yac_component_delete(component);
    }

    yac_components_reset();
    yac_xml_destroy_config_xml();
  }

  { // tests with couple_config_test_1_9.xml
    yac_xml_init_config_xml("couple_config_test_1_9.xml", "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    { // tests various interpolation methods

      struct component * components[1];
      char * component_names[2] = {"comp_1", "comp_2"};
      unsigned component_id;

      components[0] = yac_component_new(component_names[rank >> 1]);
      component_id = yac_pointer_to_unique_id(components[0]);

      double coordinates_x[5] = {0.0,1.0,2.0,3.0,4.0};
      double coordinates_y[5] = {0.0,1.0,2.0,3.0,4.0};
      size_t const num_cells[2] = {4,4};
      size_t local_start[3][2] = {{0,0},{2,0}, {0,0}};
      size_t local_count[3][2] = {{2,4},{2,4}, {4,4}};
      int with_halo = 0;
      for (size_t i = 0; i < 5; ++i) coordinates_x[i] *= YAC_RAD;
      for (size_t i = 0; i < 5; ++i) coordinates_y[i] *= YAC_RAD;

      char * grid_name[2] = {"grid1", "grid2"};
      struct basic_grid_data * grid_data =
        yac_generate_basic_grid_data_reg2d(
          coordinates_x, coordinates_y, num_cells,
          local_start[rank], local_count[rank], with_halo);
      struct basic_grid grid;
      grid.name = grid_name[rank>>1];
      grid.data = *grid_data;

      struct interp_field interp_fields[2];
      interp_fields[0].location = CORNER;
      interp_fields[0].coordinates = NULL;
      interp_fields[0].mask = NULL;
      interp_fields[1].location = CELL;
      interp_fields[1].coordinates = NULL;
      interp_fields[1].mask = NULL;

      size_t const field_count = 12;
      struct coupling_field * coupling_fields[field_count];
      char * field_name[] = {"AVG_ARITHMETIC",
                             "AVG_DIST",
                             "4NN_ARITHMETIC",
                             "4NN_DIST",
                             "4NN_GAUSS",
                             "HCSBB",
                             "RBF_4_GAUSS",
                             "SPMAP",
                             "FIXED",
                             "CONSERV_FRACAREA",
                             "CONSERV_DESTAREA",
                             "CONSERV2ND"};

      for (size_t i = 0; i < 9; ++i)
        coupling_fields[i] =
          yac_coupling_field_new(
            field_name[i], component_id, &grid, &(interp_fields[0]), 1);
      for (size_t i = 9; i < field_count; ++i)
        coupling_fields[i] =
          yac_coupling_field_new(
            field_name[i], component_id, &grid, &(interp_fields[1]), 1);

      yac_setup_couple_config(1, components, field_count, coupling_fields);

      for (size_t i = 0; i < field_count; ++i)
        yac_coupling_field_delete(coupling_fields[i]);
      yac_basic_grid_data_free(grid.data);
      free(grid_data);
      yac_component_delete(components[0]);
    }

    yac_components_reset();
    yac_xml_destroy_config_xml();
  }

  yac_finalize();
  xt_finalize();
  yac_mpi_call(MPI_Finalize(), MPI_COMM_WORLD);

  return TEST_EXIT_CODE;
}

