/**
 * @file test_couple_config_parallel2.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <mpi.h>

#include "tests.h"
#include "test_common.h"
#include "xml/config_xml.h"
#include "couple_config.h"
#include "utils.h"
#include "interp_method_check.h"
#include "yac_interface.h"
#include "yac_mpi.h"

static void check_constructor(void * user_data);

static void check_do_search(
  yac_int const * global_ids, double const (*coordinates_xyz)[3],
  size_t count, void * user_data);

static int get_num_unique_configurations(unsigned tgt_config[3][4][2],
                                         unsigned src_config[3][4][2]);

int main (void) {

  yac_mpi_call(MPI_Init(NULL, NULL), MPI_COMM_WORLD);
  yac_mpi_call(
    MPI_Comm_dup(MPI_COMM_WORLD, &yac_mpi_comm_world), MPI_COMM_WORLD);
  xt_initialize(yac_mpi_comm_world);
  int rank, size;
  yac_mpi_call(
    MPI_Comm_rank(yac_mpi_comm_world, &rank), yac_mpi_comm_world);
  yac_mpi_call(
    MPI_Comm_size(yac_mpi_comm_world, &size), yac_mpi_comm_world);

  if (size != 9) {

    PUT_ERR("ERROR: wrong number of processes\n");
    return TEST_EXIT_CODE;
  }

  char const * xml_config_filename = "couple_config_2_test_1.xml";
  char const * xml_config_file_format =
#include "test_couple_config_xml_string.h"
;

  unsigned num_components = 3;

  unsigned field_configuration[6][3][4][2];

  for (unsigned comp_rank = 0; comp_rank < 3; ++comp_rank) {
    // all fields have different configurations on all processes
    field_configuration[0][comp_rank][0][0] = 0;
    field_configuration[0][comp_rank][1][0] = 1;
    field_configuration[0][comp_rank][2][0] = 2;
    field_configuration[0][comp_rank][3][0] = 3;
    field_configuration[0][comp_rank][0][1] = 0;
    field_configuration[0][comp_rank][1][1] = 1;
    field_configuration[0][comp_rank][2][1] = 2;
    field_configuration[0][comp_rank][3][1] = 3;

    // all fields have identical configurations on all processes
    field_configuration[1][comp_rank][0][0] = 0;
    field_configuration[1][comp_rank][1][0] = 0;
    field_configuration[1][comp_rank][2][0] = 0;
    field_configuration[1][comp_rank][3][0] = 0;
    field_configuration[1][comp_rank][0][1] = 0;
    field_configuration[1][comp_rank][1][1] = 0;
    field_configuration[1][comp_rank][2][1] = 0;
    field_configuration[1][comp_rank][3][1] = 0;

    // there are only two different configurations on all processes
    field_configuration[2][comp_rank][0][0] = 0;
    field_configuration[2][comp_rank][1][0] = 0;
    field_configuration[2][comp_rank][2][0] = 1;
    field_configuration[2][comp_rank][3][0] = 1;
    field_configuration[2][comp_rank][0][1] = 0;
    field_configuration[2][comp_rank][1][1] = 0;
    field_configuration[2][comp_rank][2][1] = 1;
    field_configuration[2][comp_rank][3][1] = 1;

    // all fields have identical configurations on all processes except the
    // first one
    field_configuration[3][comp_rank][0][0] = 0;
    field_configuration[3][comp_rank][1][0] = 0;
    field_configuration[3][comp_rank][2][0] = 0;
    field_configuration[3][comp_rank][3][0] = 0;
    field_configuration[3][comp_rank][0][1] = (comp_rank == 0)?0:0;
    field_configuration[3][comp_rank][1][1] = (comp_rank == 0)?1:0;
    field_configuration[3][comp_rank][2][1] = (comp_rank == 0)?2:0;
    field_configuration[3][comp_rank][3][1] = (comp_rank == 0)?3:0;

    //
    field_configuration[4][comp_rank][0][0] = 0;
    field_configuration[4][comp_rank][1][0] = 0;
    field_configuration[4][comp_rank][2][0] = 2;
    field_configuration[4][comp_rank][3][0] = 3;
    field_configuration[4][comp_rank][0][1] = (comp_rank&1)?0:0;
    field_configuration[4][comp_rank][1][1] = (comp_rank&1)?1:0;
    field_configuration[4][comp_rank][2][1] = (comp_rank&1)?2:2;
    field_configuration[4][comp_rank][3][1] = (comp_rank&1)?2:3;

    //
    field_configuration[5][comp_rank][0][0] = 0;
    field_configuration[5][comp_rank][1][0] = 1;
    field_configuration[5][comp_rank][2][0] = 2;
    field_configuration[5][comp_rank][3][0] = 0;
    field_configuration[5][comp_rank][0][1] = 0;
    field_configuration[5][comp_rank][1][1] = 1;
    field_configuration[5][comp_rank][2][1] = 2;
    field_configuration[5][comp_rank][3][1] = 0;
  }

  char check_constructor_key[2][1024];
  char check_do_search_key[2][1024];

  sprintf(check_constructor_key[0], "check_constructor_%u_to_%u",
          (rank % num_components) + 1,(((rank % num_components) + 1) % 3) + 1);
  sprintf(check_constructor_key[1], "check_constructor_%u_to_%u",
          (rank % num_components) + 1,(((rank % num_components) + 2) % 3) + 1);

  sprintf(check_do_search_key[0], "check_do_search_%u_to_%u",
          (rank % num_components) + 1,(((rank % num_components) + 1) % 3) + 1);
  sprintf(check_do_search_key[1], "check_do_search_%u_to_%u",
          (rank % num_components) + 1,(((rank % num_components) + 2) % 3) + 1);

  int num_interpolation_method_stacks[2][2];

  for (unsigned i = 0; i < 2; ++i) {
    yac_interp_method_check_add_constructor_callback(
      check_constructor, &(num_interpolation_method_stacks[0][i]),
      check_constructor_key[i]);
    yac_interp_method_check_add_do_search_callback(
      check_do_search, &(num_interpolation_method_stacks[1][i]),
      check_do_search_key[i]);
  }

  for (unsigned tgt_config = 0; tgt_config < 6; ++tgt_config) {
    for (unsigned src_config = 0; src_config < 6; ++src_config) {

      for (unsigned i = 0; i < 2; ++i)
        for (unsigned j = 0; j < 2; ++j)
          num_interpolation_method_stacks[i][j] = 0;

      MPI_Barrier(MPI_COMM_WORLD);
      if (rank == 0) {

        // generate xml file
        FILE * xml_config_file = fopen(xml_config_filename, "w+");
        fprintf(xml_config_file, xml_config_file_format,
                field_configuration[tgt_config][0][0][0] + 1,
                field_configuration[tgt_config][0][1][0] + 1,
                field_configuration[tgt_config][0][2][0] + 1,
                field_configuration[tgt_config][0][3][0] + 1,
                field_configuration[src_config][0][0][0] + 1,
                field_configuration[src_config][0][1][0] + 1,
                field_configuration[src_config][0][2][0] + 1,
                field_configuration[src_config][0][3][0] + 1,
                field_configuration[tgt_config][1][0][0] + 1,
                field_configuration[tgt_config][1][1][0] + 1,
                field_configuration[tgt_config][1][2][0] + 1,
                field_configuration[tgt_config][1][3][0] + 1,
                field_configuration[src_config][1][0][0] + 1,
                field_configuration[src_config][1][1][0] + 1,
                field_configuration[src_config][1][2][0] + 1,
                field_configuration[src_config][1][3][0] + 1,
                field_configuration[tgt_config][2][0][0] + 1,
                field_configuration[tgt_config][2][1][0] + 1,
                field_configuration[tgt_config][2][2][0] + 1,
                field_configuration[tgt_config][2][3][0] + 1,
                field_configuration[src_config][2][0][0] + 1,
                field_configuration[src_config][2][1][0] + 1,
                field_configuration[src_config][2][2][0] + 1,
                field_configuration[src_config][2][3][0] + 1);
        fclose(xml_config_file);
      }

      MPI_Barrier(MPI_COMM_WORLD);

      yac_xml_init_config_xml(xml_config_filename, "coupling.xsd");
      yac_cdef_datetime ( NULL, NULL) ;

      char * component_names[3] = {"comp_1", "comp_2", "comp_3"};
      struct component * component;
      unsigned component_id;
      MPI_Comm component_comm;
      int comp_rank;

      component = yac_component_new(component_names[rank % num_components]);
      component_id = yac_pointer_to_unique_id(component);
      component_comm = yac_component_get_communicator(component);
      MPI_Comm_rank(component_comm, &comp_rank);

      size_t num_vertices[2] = {2,2};
      int cyclic[2] = {0,0};
      double coordinates_x[2];
      double coordinates_y[4][2] = {{0,1},{1,2},{2,3},{3,4}};

      coordinates_x[0] = comp_rank;
      coordinates_x[1] = comp_rank + 1;

      yac_int global_cell_ids[1] = {comp_rank};
      yac_int global_corner_ids[4] =
        {4 * comp_rank + 0, 4 * comp_rank + 1,
         4 * comp_rank + 2, 4 * comp_rank + 3};
      int cell_core_mask[1] = {1};
      int corner_core_mask[4] = {1,1,1,1};

      struct basic_grid grid[2][4];
      char grid_names[2][4][1024];

      for (unsigned i = 0; i < 4; ++i) {
        for (unsigned j = 0; j < 2; ++j) {
          sprintf(grid_names[j][i], "grid_%u_%s", i+1, (j == 0)?"in":"out");
          grid[j][i].name = grid_names[j][i];
          grid[j][i].data =
            yac_generate_basic_grid_data_reg_2d_deg(
              num_vertices, cyclic, coordinates_x,
              coordinates_y[
                field_configuration[
                  (j == 0)?tgt_config:src_config][comp_rank][i][0]]);
          grid[j][i].data.cell_ids = TO_POINTER(global_cell_ids);
          grid[j][i].data.vertex_ids = TO_POINTER(global_corner_ids);
          grid[j][i].data.core_cell_mask = TO_POINTER(cell_core_mask);
          grid[j][i].data.core_vertex_mask = TO_POINTER(corner_core_mask);
        }
      }

      struct interp_field interp_fields[4];
      for (unsigned i = 0; i < 4; ++i) {
        interp_fields[i].location = CORNER;
        interp_fields[i].coordinates = grid[0][i].data.vertex_coordinates;
        interp_fields[i].mask = NULL;
      }

      char field_name_in[8][1024], field_name_out[8][1024];

      {
        unsigned count = 0;

        for (unsigned i = 0; i < 3; ++i) {

          if ((rank % num_components) == i) continue;

          for (unsigned j = 1; j <= 4; ++j) {
            sprintf(field_name_out[count], "transient_from_%u_to_%u_%u",
                    (rank % num_components)+1, i+1, j);
            sprintf(field_name_in[count], "transient_from_%u_to_%u_%u", i+1,
                    (rank % num_components)+1, j);
            ++count;
          }
        }
      }

      struct coupling_field * coupling_fields[2][8];

      for (unsigned i = 0; i < 8; ++i) {
        coupling_fields[0][i] =
          yac_coupling_field_new(
            field_name_in[i], component_id,
            &grid[0][field_configuration[tgt_config][comp_rank][i&3][0]],
            &interp_fields[field_configuration[tgt_config][comp_rank][i&3][1]],
            1);
        coupling_fields[1][i] =
          yac_coupling_field_new(
            field_name_out[i], component_id,
            &grid[1][field_configuration[src_config][comp_rank][i&3][0]],
            &interp_fields[field_configuration[src_config][comp_rank][i&3][1]],
            1);
      }

      yac_setup_couple_config(1, &component, 16, &(coupling_fields[0][0]));

      unsigned ref_num_unique_configs =
        get_num_unique_configurations(field_configuration[tgt_config],
                                      field_configuration[src_config]);

      for (unsigned i = 0; i < 2; ++i)
        for (unsigned j = 0; j < 2; ++j)
          if (num_interpolation_method_stacks[i][j] != ref_num_unique_configs)
            PUT_ERR("ERROR: interpolation was not correctly reused\n");

      for (unsigned i = 0; i < 2; ++i)
        for (unsigned j = 0; j < 8; ++j)
          yac_coupling_field_delete(coupling_fields[i][j]);
      for (unsigned i = 0; i < 4; ++i)
        for (unsigned j = 0; j < 2; ++j)
          yac_basic_grid_data_free(grid[j][i].data);
      yac_component_delete(component);

      yac_xml_destroy_config_xml();

      yac_components_reset();
    }
  }

  yac_finalize();

  yac_mpi_call(MPI_Finalize(), MPI_COMM_WORLD);

  if (rank == 0) unlink(xml_config_filename);

  return TEST_EXIT_CODE;
}

static void check_constructor(void * user_data) {

  ++*(int*)user_data;

  return;
}

static void check_do_search(
  yac_int const * global_ids, double const (*coordinates_xyz)[3],
  size_t num_points, void * user_data) {

  UNUSED(global_ids);
  UNUSED(coordinates_xyz);
  UNUSED(num_points);

  ++*(int*)user_data;

  return;
}

static int get_num_unique_configurations(unsigned tgt_config[3][4][2],
                                         unsigned src_config[3][4][2]) {

  unsigned num_configs = 0;

  for (unsigned i = 0; i < 4; ++i) {

    unsigned k;

    for (k = i + 1; k < 4; ++k) {

      int differs = 0;

      for (unsigned j = 0; j < 3; ++j)
        differs |= (tgt_config[j][i][1] != tgt_config[j][k][1]) ||
                   (src_config[j][i][1] != src_config[j][k][1]);

      if (!differs) break;
    }

    num_configs += k == 4;
  }

  return num_configs;
}

