/**
 * @file test_couple_config_parallel3.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <mpi.h>
#include <yaxt.h>

#include "tests.h"
#include "test_common.h"
#include "xml/config_xml.h"
#include "couple_config.h"
#include "utils.h"
#include "yac_interface.h"
#include "yac_mpi.h"
#include "yac_mpi.h"

char * str_logical[2] = {"true", "false"};

int main (void) {

  yac_mpi_call(MPI_Init(NULL, NULL), MPI_COMM_WORLD);
  yac_mpi_call(
    MPI_Comm_dup(MPI_COMM_WORLD, &yac_mpi_comm_world), MPI_COMM_WORLD);
  xt_initialize(yac_mpi_comm_world);
  int rank, size;
  yac_mpi_call(
    MPI_Comm_rank(yac_mpi_comm_world, &rank), yac_mpi_comm_world);
  yac_mpi_call(
    MPI_Comm_size(yac_mpi_comm_world, &size), yac_mpi_comm_world);

  if (size != 9) {

    PUT_ERR("ERROR: wrong number of processes\n");
    return TEST_EXIT_CODE;
  }

  char const * xml_config_filename = "couple_config_3_test.xml";
  char const * xml_config_file_format =
#include "test_couple_config_3_xml_string.h"
;

  for (unsigned n = 0; n < 1 << 6; ++n) {

    MPI_Barrier(MPI_COMM_WORLD);
    if (rank == 0) {
      // generate xml file
      FILE * xml_config_file = fopen(xml_config_filename, "w+");
      fprintf(xml_config_file, xml_config_file_format,
              str_logical[!(n & (1 << 0))], str_logical[!(n & (1 << 1))],
              str_logical[!(n & (1 << 2))], str_logical[!(n & (1 << 3))],
              str_logical[!(n & (1 << 4))], str_logical[!(n & (1 << 5))]);
      fclose(xml_config_file);
    }
    MPI_Barrier(MPI_COMM_WORLD);

    yac_xml_init_config_xml(xml_config_filename, "coupling.xsd");
    yac_cdef_datetime ( NULL, NULL) ;

    char * component_names[3] = {"comp_1", "comp_2", "comp_3"};
    char const * local_component_names[2];
    struct component * components[2];
    unsigned component_ids[2];
    MPI_Comm component_comms[2];
    int comp_ranks[2];
    int comp_idx[3][2] = {{0,1},{0,2},{1,2}};

    for (unsigned i = 0; i < 2; ++i)
      local_component_names[i] = component_names[comp_idx[rank/3][i]];

    yac_components_new(local_component_names, 2, components);

    for (unsigned i = 0; i < 2; ++i) {
      component_ids[i] = yac_pointer_to_unique_id(components[i]);
      component_comms[i] = yac_component_get_communicator(components[i]);
      MPI_Comm_rank(component_comms[i], comp_ranks + i);
    }

    struct basic_grid grid[2];
    char grid_names[2][1024];
    size_t num_vertices[2] = {2,2};
    int cyclic[2] = {0,0};
    double coordinates_x[2][2];
    double coordinates_y[2] = {0,1};
    yac_int global_cell_ids[2][1];
    yac_int global_corner_ids[2][4];
    yac_int global_edge_ids[2][4];
    int cell_core_mask = {1};
    int corner_core_mask[4] = {1,1,1,1};
    int edge_core_mask[4] = {1,1,1,1};
    int ref_global_edge_ids[6][4] = {{0,3,13,1},{2,5,14,3},{4,7,15,5},
                                     {6,9,16,7},{8,11,17,9},{10,12,18,11}};

    for (unsigned i = 0; i < 2; ++i) {
      sprintf(grid_names[i], "grid_%d", comp_idx[rank/3][i] + 1);
      coordinates_x[i][0] = comp_ranks[i];
      coordinates_x[i][1] = comp_ranks[i] + 1;
      global_cell_ids[i][0] = comp_ranks[i];
      global_corner_ids[i][0] = comp_ranks[i] + 0;
      global_corner_ids[i][1] = comp_ranks[i] + 1;
      global_corner_ids[i][2] = comp_ranks[i] + 7;
      global_corner_ids[i][3] = comp_ranks[i] + 8;
      global_edge_ids[i][0] = ref_global_edge_ids[comp_ranks[i]][0];
      global_edge_ids[i][1] = ref_global_edge_ids[comp_ranks[i]][1];
      global_edge_ids[i][2] = ref_global_edge_ids[comp_ranks[i]][2];
      global_edge_ids[i][3] = ref_global_edge_ids[comp_ranks[i]][3];
      grid[i].name = grid_names[i];
      grid[i].data =
        yac_generate_basic_grid_data_reg_2d_deg(
          num_vertices, cyclic, coordinates_x[i], coordinates_y);
      grid[i].data.cell_ids = TO_POINTER(global_cell_ids[i]);
      grid[i].data.vertex_ids = TO_POINTER(global_corner_ids[i]);
      grid[i].data.edge_ids = TO_POINTER(global_edge_ids[i]);
      grid[i].data.core_cell_mask = TO_POINTER(cell_core_mask);
      grid[i].data.core_vertex_mask = TO_POINTER(corner_core_mask);
      grid[i].data.core_edge_mask = TO_POINTER(edge_core_mask);
    }

    struct interp_field interp_fields[2];

    for (unsigned i = 0; i < 2; ++i) {
      interp_fields[i].location = CORNER;
      interp_fields[i].coordinates = grid[i].data.vertex_coordinates;
      interp_fields[i].mask = NULL;
    }

    char field_name[2][2][2][1024];

    for (unsigned i = 0; i < 2; ++i) {
      int curr_comp_idx = comp_idx[rank/3][i];
      for (unsigned j = 0; j < 2; ++j) {
        int remote_comp_idx = (3 + curr_comp_idx + j + 1) % 3;
        sprintf(field_name[i][j][0], "transient_from_%d_to_%d",
                curr_comp_idx + 1, remote_comp_idx + 1);
        sprintf(field_name[i][j][1], "transient_from_%d_to_%d",
                remote_comp_idx + 1, curr_comp_idx + 1);
      }
    }

    struct coupling_field * coupling_fields[2][2][2];

    for (unsigned i = 0; i < 2; ++i)
      for (unsigned j = 0; j < 2; ++j)
        for (unsigned k = 0; k < 2; ++k)
          coupling_fields[i][j][k] =
            yac_coupling_field_new(field_name[i][j][k], component_ids[i],
                                   &grid[i], &interp_fields[i], 1);

    yac_setup_couple_config(2, components, 8, &(coupling_fields[0][0][0]));

    // unsigned ref_num_unique_configs =
      // get_num_unique_configurations(field_configuration[tgt_config],
                                    // field_configuration[src_config]);

    // for (unsigned i = 0; i < 2; ++i)
      // for (unsigned j = 0; j < 2; ++j)
        // if (num_interpolation_method_stacks[i][j] != ref_num_unique_configs)
          // PUT_ERR("ERROR: interpolation was not correctly reused\n");

    for (unsigned i = 0; i < 2; ++i)
      for (unsigned j = 0; j < 2; ++j)
        for (unsigned k = 0; k < 2; ++k)
          yac_coupling_field_delete(coupling_fields[i][j][k]);

    for (unsigned i = 0; i < 2; ++i) {
      yac_basic_grid_data_free(grid[i].data);
      yac_component_delete(components[i]);
    }

    yac_xml_destroy_config_xml();

    yac_components_reset();
  }

  yac_finalize();

  yac_mpi_call(MPI_Finalize(), MPI_COMM_WORLD);

  if (rank == 0) unlink(xml_config_filename);

  return TEST_EXIT_CODE;
}

// static int get_num_unique_configurations(unsigned tgt_config[3][4],
                                         // unsigned src_config[3][4]) {

  // unsigned num_configs = 0;

  // for (unsigned i = 0; i < 4; ++i) {

    // unsigned k;

    // for (k = i + 1; k < 4; ++k) {

      // int differs = 0;

      // for (unsigned j = 0; j < 3; ++j)
        // differs |= (tgt_config[j][i] != tgt_config[j][k]) ||
                   // (src_config[j][i] != src_config[j][k]);

      // if (!differs) break;
    // }

    // num_configs += k == 4;
  // }

  // return num_configs;
// }

