/**
 * @file test_interp_method_nnn_parallel2.c
 *
 * @copyright Copyright  (C)  2020 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "tests.h"
#include "interp_method.h"
#include "interp_method_nnn.h"
#include "dist_grid_utils.h"
#include "yac_mpi.h"

#if defined(YAC_NETCDF_ENABLED)

#include <mpi.h>
#include <yaxt.h>
#include <netcdf.h>

int const grid_id_src = 1;
int const grid_id_tgt = 2;

static void target_main(MPI_Comm global_comm, MPI_Comm target_comm);
static void source_main(MPI_Comm global_comm, MPI_Comm source_comm);

int main (void) {

  MPI_Init(NULL, NULL);

  xt_initialize(MPI_COMM_WORLD);

  int comm_rank, comm_size;
  yac_mpi_call(MPI_Comm_rank(MPI_COMM_WORLD, &comm_rank), MPI_COMM_WORLD);
  yac_mpi_call(MPI_Comm_size(MPI_COMM_WORLD, &comm_size), MPI_COMM_WORLD);
  MPI_Barrier(MPI_COMM_WORLD);

  if (comm_size != 5) {
    PUT_ERR("ERROR: wrong number of processes");
    xt_finalize();
    MPI_Finalize();
    return TEST_EXIT_CODE;
  }

  int tgt_flag = comm_rank < 1;

  MPI_Comm split_comm;
  yac_mpi_call(
    MPI_Comm_split(
      MPI_COMM_WORLD, tgt_flag, 0, &split_comm), MPI_COMM_WORLD);

  if (tgt_flag) target_main(MPI_COMM_WORLD, split_comm);
  else          source_main(MPI_COMM_WORLD, split_comm);

  yac_mpi_call(MPI_Comm_free(&split_comm), MPI_COMM_WORLD);

  xt_finalize();

  MPI_Finalize();

  return TEST_EXIT_CODE;
}

static void source_main(MPI_Comm global_comm,
                        MPI_Comm source_comm) {

  // 1 and 5 nearest neighbours per target point with fixed value fallback

  // corner and cell ids for a 7 x 7 grid (x == target point position)

  // 56-----57-----58-----59-----60-----61-----62-----63
  //  |     x|     x|     x|     x|     x|     x|     x|
  //  |  42  |  43  |  44  |  45  |  46  |  47  |  48  |
  //  |      |      |      |      |      |      |      |
  // 48-----49-----50-----51-----52-----53-----54-----55
  //  |     x|     x|     x|     x|     x|     x|     x|
  //  |  35  |  36  |  37  |  38  |  39  |  40  |  41  |
  //  |      |      |      |      |      |      |      |
  // 40-----41-----42-----43-----44-----45-----46-----47
  //  |     x|     x|     x|     x|     x|     x|     x|
  //  |  28  |  29  |  30  |  31  |  32  |  33  |  34  |
  //  |      |      |      |      |      |      |      |
  // 32-----33-----34-----35-----36-----37-----38-----39
  //  |     x|     x|     x|     x|     x|     x|     x|
  //  |  21  |  22  |  23  |  24  |  25  |  26  |  27  |
  //  |      |      |      |      |      |      |      |
  // 24-----25-----26-----27-----28-----29-----30-----31
  //  |     x|     x|     x|     x|     x|     x|     x|
  //  |  14  |  15  |  16  |  17  |  18  |  19  |  20  |
  //  |      |      |      |      |      |      |      |
  // 16-----17-----18-----19-----20-----21-----22-----23
  //  |     x|     x|     x|     x|     x|     x|     x|
  //  |  07  |  08  |  09  |  10  |  11  |  12  |  13  |
  //  |      |      |      |      |      |      |      |
  // 08-----09-----10-----11-----12-----13-----14-----15
  //  |     x|     x|     x|     x|     x|     x|     x|
  //  |  00  |  01  |  02  |  03  |  04  |  05  |  06  |
  //  |      |      |      |      |      |      |      |
  // 00-----01-----02-----03-----04-----05-----06-----07
  //
  // the grid is distributed among the processes as follows:
  // (index == process)
  //
  // 3---3---3---3---3---3---3---3
  // | 3 | 3 | 3 | 3 | 3 | 3 | 3 |
  // 3---3---3---3---3---3---3---3
  // | 3 | 3 | 3 | 3 | 3 | 3 | 3 |
  // 3---3---3---1---2---2---3---3
  // | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
  // 1---1---1---2---2---2---2---2
  // | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
  // 1---1---1---1---2---2---2---2
  // | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
  // 1---1---1---0---0---0---2---2
  // | 0 | 0 | 0 | 0 | 0 | 0 | 0 |
  // 0---0---0---0---0---0---0---0
  // | 0 | 0 | 0 | 0 | 0 | 0 | 0 |
  // 0---0---0---0---0---0---0---0
  //
  // the source mask looks as follows (# == masked point)
  //
  // +---+---#---#---#---#---#---#
  // |   |   |   |   |   |   |   |
  // +---+---#---#---#---#---#---#
  // |   |   |   |   |   |   |   |
  // #---#---#---#---#---#---#---#
  // |   |   |   |   |   |   |   |
  // #---#---#---#---#---#---#---#
  // |   |   |   |   |   |   |   |
  // #---#---#---#---#---#---#---#
  // |   |   |   |   |   |   |   |
  // #---#---#---#---#---#---#---#
  // |   |   |   |   |   |   |   |
  // +---#---#---#---#---#---#---+
  // |   |   |   |   |   |   |   |
  // +---+---#---#---#---#---+---+
  //
  // the target mask looks as follows (# == masked point)
  //
  // +---+---+---+---+---+---+---+
  // |   |   |   |   |   |   |   |
  // +---+---+---+---+---+---+---+
  // |   |   |   |   |   |   |   |
  // +---+---#---#---#---#---+---+
  // |   |   |   |   |   |   |   |
  // +---+---#---#---#---#---+---+
  // |   |   |   |   |   |   |   |
  // +---+---#---#---#---#---+---+
  // |   |   |   |   |   |   |   |
  // +---+---#---#---#---#---+---+
  // |   |   |   |   |   |   |   |
  // +---+---+---+---+---+---+---+
  // |   |   |   |   |   |   |   |
  // +---+---+---+---+---+---+---+
  //--------------------
  // set up global search
  //--------------------

  int my_source_rank;
  yac_mpi_call(MPI_Comm_rank(source_comm, &my_source_rank), source_comm);

  double coordinates_x[] = {0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0};
  double coordinates_y[] = {0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0};
  size_t const num_cells[2] = {7,7};
  size_t local_start[4][2] = {{0,0},{0,2},{3,2},{0,5}};
  size_t local_count[4][2] = {{7,2},{3,3},{4,3},{7,2}};
  int with_halo = 0;
  int global_corner_mask[8][8] = {
    {1,1,0,0,0,0,1,1},
    {1,0,0,0,0,0,0,1},
    {0,0,0,0,0,0,0,0},
    {0,0,0,0,0,0,0,0},
    {0,0,0,0,0,0,0,0},
    {0,0,0,0,0,0,0,0},
    {1,1,0,0,0,0,0,0},
    {1,1,0,0,0,0,0,0}};
  for (size_t i = 0; i <= num_cells[0]; ++i) coordinates_x[i] *= YAC_RAD;
  for (size_t i = 0; i <= num_cells[1]; ++i) coordinates_y[i] *= YAC_RAD;

  struct basic_grid_data * grid_data =
    yac_generate_basic_grid_data_reg2d(
      coordinates_x, coordinates_y, num_cells,
      local_start[my_source_rank], local_count[my_source_rank], with_halo);

  struct dist_grid_pair * grid_pair =
    yac_dist_grid_pair_new(grid_data, grid_id_src,
                           &basic_grid_data_empty, grid_id_tgt, MPI_COMM_WORLD);

  int * local_corner_mask =
    xmalloc(grid_data->num_vertices * sizeof(*local_corner_mask));
  for (size_t i = 0; i < grid_data->num_vertices; ++i)
    local_corner_mask[i] =
      ((int*)(&(global_corner_mask[0][0])))[grid_data->vertex_ids[i]];

  struct interp_field src_fields[] =
    {{.location = CORNER, .coordinates = NULL, .mask = local_corner_mask}};
  size_t num_src_fields = sizeof(src_fields) / sizeof(src_fields[0]);
  struct interp_field tgt_field =
    {.location = CELL, .coordinates = NULL, .mask = NULL};

  struct interp_grid * interp_grid =
    yac_interp_grid_new(grid_pair, grid_id_src, grid_id_tgt,
                        num_src_fields, src_fields, tgt_field);

  struct interp_method * method_stack[2] = {
    yac_interp_method_nnn_new(
      (struct yac_nnn_config){.type = NNN_AVG, .n = 4}), NULL};

  struct interp_weights * weights =
    yac_interp_method_do_search(method_stack, interp_grid);

  yac_interp_method_delete(method_stack);

  enum interp_weights_reorder_type reorder_type[2] =
    {MAPPING_ON_SRC, MAPPING_ON_TGT};

  for (size_t i = 0; i < 2; ++i) {

    struct interpolation * interpolation =
      yac_interp_weights_get_interpolation(weights, reorder_type[i], 1);

    // check generated interpolation
    {
      double * src_field = NULL;
      double ** src_fields = &src_field;

      src_field = xmalloc(grid_data->num_vertices * sizeof(*src_field));
      for (size_t i = 0; i < grid_data->num_vertices; ++i)
        src_field[i] = (double)(grid_data->vertex_ids[i]);

      yac_interpolation_execute(interpolation, &src_fields, NULL);

      free(src_field);
    }

    yac_interpolation_delete(interpolation);
  }

  yac_basic_grid_data_delete(grid_data);
  yac_interp_weights_delete(weights);
  yac_interp_grid_delete(interp_grid);
  yac_dist_grid_pair_delete(grid_pair);
  free(local_corner_mask);
}

static void target_main(MPI_Comm global_comm,
                        MPI_Comm target_comm) {

  double coordinates_x[] = {0.0,1.0,2.0,3.0,4.0,5.0,6.0,7.0};
  double coordinates_y[] = {0.0,1.0,2.0,3.0,4.0,5.0,6.0,7.0};
  double cell_coordinates_x[] = {0.75,1.75,2.75,3.75,4.75,5.75,6.75};
  double cell_coordinates_y[] = {0.75,1.75,2.75,3.75,4.75,5.75,6.75};
  double cell_coordinates[49][3];
  size_t const num_cells[2] = {7,7};
  size_t local_start[2] = {0,0};
  size_t local_count[2] = {7,7};
  int with_halo = 0;
  int global_cell_mask[7][7] = {
    {1,1,1,1,1,1,1},
    {1,1,1,1,1,1,1},
    {1,1,0,0,0,1,1},
    {1,1,0,0,0,1,1},
    {1,1,0,0,0,1,1},
    {1,1,1,1,1,1,1},
    {1,1,1,1,1,1,1}};
  for (size_t i = 0; i <= num_cells[0]; ++i) coordinates_x[i] *= YAC_RAD;
  for (size_t i = 0; i <= num_cells[1]; ++i) coordinates_y[i] *= YAC_RAD;
  for (size_t i = 0, k = 0; i < num_cells[1]; ++i)
    for (size_t j = 0; j < num_cells[0]; ++j, ++k)
      LLtoXYZ_deg(
        cell_coordinates_x[j], cell_coordinates_y[i], cell_coordinates[k]);

  struct basic_grid_data * grid_data =
    yac_generate_basic_grid_data_reg2d(
      coordinates_x, coordinates_y, num_cells,
      local_start, local_count, with_halo);

  struct dist_grid_pair * grid_pair =
    yac_dist_grid_pair_new(grid_data, grid_id_tgt,
                           &basic_grid_data_empty, grid_id_src, MPI_COMM_WORLD);

  int * local_cell_mask =
    xmalloc(grid_data->num_cells * sizeof(*local_cell_mask));
  for (size_t i = 0; i < grid_data->num_cells; ++i)
    local_cell_mask[i] =
      ((int*)(&(global_cell_mask[0][0])))[grid_data->cell_ids[i]];

  struct interp_field src_fields[] =
    {{.location = CORNER, .coordinates = NULL, .mask = NULL}};
  size_t num_src_fields = sizeof(src_fields) / sizeof(src_fields[0]);
  struct interp_field tgt_field =
    {.location = CELL,
     .coordinates = (coordinate_pointer)&(cell_coordinates[0][0]),
     .mask = local_cell_mask};

  struct interp_grid * interp_grid =
    yac_interp_grid_new(grid_pair, grid_id_src, grid_id_tgt,
                        num_src_fields, src_fields, tgt_field);

  struct interp_method * method_stack[2] = {
    yac_interp_method_nnn_new(
      (struct yac_nnn_config){.type = NNN_AVG, .n = 4}), NULL};

  struct interp_weights * weights =
    yac_interp_method_do_search(method_stack, interp_grid);

  yac_interp_method_delete(method_stack);

  enum interp_weights_reorder_type reorder_type[2] =
    {MAPPING_ON_SRC, MAPPING_ON_TGT};

  for (size_t i = 0; i < 2; ++i) {

    struct interpolation * interpolation =
      yac_interp_weights_get_interpolation(weights, reorder_type[i], 1);

    // check generated interpolation
    {
      double * tgt_field = NULL;
      double ref_tgt_field[49] =
        {(0+1+8+49)/4.0,(0+1+8+6)/4.0,(0+1+8+6)/4.0,(1+6+7+15)/4.0,
           (1+6+7+15)/4.0,(1+6+7+15)/4.0,(1+6+7+15)/4.0,
         (0+1+8+49)/4.0,(0+1+8+49)/4.0,(0+1+8+6)/4.0,(1+6+7+15)/4.0,
           (1+6+7+15)/4.0,(1+6+7+15)/4.0,(1+6+7+15)/4.0,
         (8+1+0+49)/4.0,(8+1+0+49)/4.0,-1,-1,
           -1,(1+6+7+15)/4.0,(1+6+7+15)/4.0,
         (48+49+8+57)/4.0,(48+49+8+57)/4.0,-1,-1,
           -1,(6+7+15+49)/4.0,(6+7+15+49)/4.0,
         (48+49+56+57)/4.0,(48+49+56+57)/4.0,-1,-1,
           -1,(6+7+15+49)/4.0,(6+7+15+49)/4.0,
         (48+49+56+57)/4.0,(48+49+56+57)/4.0,(48+49+56+57)/4.0,(48+49+56+57)/4.0,
           (48+49+56+57)/4.0,(48+49+15+57)/4.0,(6+7+15+49)/4.0,
         (48+49+56+57)/4.0,(48+49+56+57)/4.0,(48+49+56+57)/4.0,(48+49+56+57)/4.0,
           (48+49+56+57)/4.0,(48+49+56+57)/4.0,(49+57+56+15)/4.0};

      tgt_field = xmalloc(grid_data->num_cells * sizeof(*tgt_field));

      for (size_t j = 0; j < grid_data->num_cells; ++j) tgt_field[j] = -1.0;

      yac_interpolation_execute(interpolation, NULL, &tgt_field);

      for (size_t j = 0; j < grid_data->num_cells; ++j)
        if (fabs(ref_tgt_field[j] - tgt_field[j]) > 1e-9)
          PUT_ERR("wrong interpolation result");

      free(tgt_field);
    }

    yac_interpolation_delete(interpolation);
  }

  yac_basic_grid_data_delete(grid_data);
  yac_interp_weights_delete(weights);
  yac_interp_grid_delete(interp_grid);
  yac_dist_grid_pair_delete(grid_pair);
  free(local_cell_mask);
}

#else
int main(void) {return EXIT_SKIP_TEST;}
#endif // YAC_NETCDF_ENABLED
