//#define VERBOSE

/**
 * @file test_partial_areas.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "grid.h"
#include "clipping.h"
#include "utils.h"
#include "area.h"
#include "tests.h"
#include "geometry.h"
#include "test_common.h"

static int compare_area(double a, double b);

int main (void) {

  /* For the test at the Equator we cannot get better than 1.0e-11, while
     for the Pole test we are somewhat better.

    tolerance test for the deviation from 1 of  the sum of weights: 1.0e-11

  */

  double area[1024];
  double tgt_area;
  double weight[100];
  double weight_sum;

  double const epsilon = 1.0e-10; // relative precision

  struct grid_cell *SourceCell = xmalloc(256 * sizeof(*SourceCell));
  struct grid_cell  TargetCell;

  size_t nSourceCells = 4;

  enum yac_edge_type gc_edges[] = {GREAT_CIRCLE, GREAT_CIRCLE,
                                   GREAT_CIRCLE, GREAT_CIRCLE};
  enum yac_edge_type latlon_edges[4] = {LAT_CIRCLE, LON_CIRCLE,
                                        LAT_CIRCLE, LON_CIRCLE};

#ifdef VERBOSE
  printf (" ---------- test 1.) ----------- \n");
#endif

  /* 1.) 4 source elements overlapping a target cell

    Target Cell

              0.0 (lon)  [1]
              0.7 (lat)
              / \
             /   \
   -0.5 [2] /     \  0.5 [0]
    0.0     \     /  0.0
             \   /
              \ /
              0.0  [3]
             -0.5
  */
  TargetCell =
    generate_cell_deg((double[]){0.5, 0.0, -0.5,  0.0},
                      (double[]){0.0, 0.7,  0.0, -0.5}, gc_edges, 4);

  /* source cell test data

         0.0          1.0 (lon)
         1.8          1.8 (lat)
   ------------------
           |        |
       1   |    0   |
           |        |
           |        |  1.0
   ------------------  0.6
         0.0        |
         0.6        |
                    |
       2   |    3   |
           |        |
   ------------------  1.0
                      -0.6
  */

  SourceCell[0] =
    generate_cell_deg((double[]){1.0, 0.0, 0.0, 1.0},
                      (double[]){1.8, 1.8, 0.6, 0.6}, gc_edges, 4);
  SourceCell[1] =
    generate_cell_deg((double[]){0.0, -1.0, -1.0, 0.0},
                      (double[]){1.8,  1.8,  0.6, 0.6}, gc_edges, 4);
  SourceCell[2] =
    generate_cell_deg((double[]){0.0, -1.0, -1.0,  0.0},
                      (double[]){0.6,  0.6, -0.6, -0.6}, gc_edges, 4);
  SourceCell[3] =
    generate_cell_deg((double[]){1.0, 0.0,  0.0,  1.0},
                      (double[]){0.6, 0.6, -0.6, -0.6}, gc_edges, 4);

  yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

  tgt_area = yac_huiliers_area(TargetCell);
  for (size_t n = 0; n < nSourceCells; ++n)
    weight[n] = area[n] / tgt_area;

  for (size_t n = 0; n < nSourceCells; ++n) yac_free_grid_cell(SourceCell + n);
  yac_free_grid_cell(&TargetCell);

  weight_sum = 0.0;
  for ( size_t n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 1.) weight %lf\n", weight_sum );
    PUT_ERR("test 1.) weight deviates from expected value of 1.0!\n");
  }

  yac_correct_weights ( nSourceCells, weight );

  weight_sum = 0.0;
  for ( size_t n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 1.) weight %lf\n", weight_sum );
    PUT_ERR("test 1.) weight deviates from expected value of 1.0!\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 2.) ----------- \n");
#endif

  /* 2.) Next test is over the pole

     Target cell located over the pole o with
     vertices x and source cells 0 to 3 arranged as


 135.0         90.0       45.0 (lon)
               89.0       88.5 (lat)
         ------------------
        |        |        |
        |        |        |
        |       1x        |
        |        |    0   |
-180.0   ---x----o----x----  0.0        pole @ -180.0; 90.0
        |   2    |        |
        |        |        |
        |       3x        |
        |        |        |
        |        |        |
-135.0   ------------------ -45.0
               -90.0         -0.6

  */

  TargetCell =
    generate_cell_deg((double[]){ 0.0, 90.0, -180.0, -90.0},
                      (double[]){89.5, 89.5,   89.5,  89.5}, gc_edges, 4);

  SourceCell[0] =
    generate_cell_deg((double[]){45.0, 90.0, -180.0,  0.0},
                      (double[]){88.5, 89.0,   90.0, 89.0}, gc_edges, 4);
  SourceCell[1] =
    generate_cell_deg((double[]){90.0, 135.0, -180.0, -180.0},
                      (double[]){89.0,  88.5,   89.0,   90.0}, gc_edges, 4);
  SourceCell[2] =
    generate_cell_deg((double[]){-180.0, -180.0, -135.0, -90.0},
                      (double[]){  90.0,   89.0,   88.5,  89.0}, gc_edges, 4);
  SourceCell[3] =
    generate_cell_deg((double[]){ 0.0, -180.0, -90.0, -45.0},
                      (double[]){89.0,   90.0,  89.0,  88.5}, gc_edges, 4);

  yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

  tgt_area = yac_huiliers_area(TargetCell);
  for (size_t n = 0; n < nSourceCells; ++n)
    weight[n] = area[n] / tgt_area;

  for (size_t n = 0; n < nSourceCells; n++) yac_free_grid_cell(SourceCell + n);
  yac_free_grid_cell(&TargetCell);

  weight_sum = 0.0;
  for ( size_t n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 2.) weight %lf\n", weight_sum );
    PUT_ERR("test 2.) weight deviates from expected value of 1.0!\n");
  }

  yac_correct_weights ( nSourceCells, weight );

  weight_sum = 0.0;
  for ( size_t n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 2.) weight %lf\n", weight_sum );
    PUT_ERR("test 2.) weight deviates from expected value of 1.0!\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 3.) ----------- \n");
#endif

  nSourceCells = 1;

  TargetCell =
    generate_cell_deg((double[]){-180.0, -150.0, -150.0, -180.0},
                      (double[]){ -90.0,  -90.0,  -60.0,  -60.0}, gc_edges, 4);
  SourceCell[0] =
    generate_cell_deg((double[]){-180.0, -150.0, -150.0, -180.0},
                      (double[]){ -90.0,  -90.0,  -60.0,  -60.0}, gc_edges, 4);

  yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

  tgt_area = yac_huiliers_area(TargetCell);
  for (size_t n = 0; n < nSourceCells; ++n)
    weight[n] = area[n] / tgt_area;

  for (size_t n = 0; n < nSourceCells; n++) yac_free_grid_cell(SourceCell + n);
  yac_free_grid_cell(&TargetCell);

  weight_sum = 0.0;
  for ( size_t n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 3.) weight %lf\n", weight_sum );
    PUT_ERR("test 3.) weight deviates from expected value of 1.0!\n");
  }

  yac_correct_weights ( nSourceCells, weight );

  weight_sum = 0.0;
  for ( size_t n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 3.) weight %lf\n", weight_sum );
    PUT_ERR("test 3.) weight deviates from expected value of 1.0!\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 4.) ----------- \n");
#endif

  TargetCell =
    generate_cell_deg((double[]){-180.0, -150.0, -150.0, -180.0},
                      (double[]){  60.0,   60.0,   90.0,   90.0}, gc_edges, 4);

  SourceCell[0] =
    generate_cell_deg((double[]){-180.0, -150.0, -150.0, -180.0},
                      (double[]){  60.0,   60.0,   90.0,   90.0}, gc_edges, 4);

  yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

  tgt_area = yac_huiliers_area(TargetCell);
  for (size_t n = 0; n < nSourceCells; ++n)
    weight[n] = area[n] / tgt_area;

  for (size_t n = 0; n < nSourceCells; n++) yac_free_grid_cell(SourceCell + n);
  yac_free_grid_cell(&TargetCell);

  weight_sum = 0.0;
  for ( size_t n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 4.) weight %lf\n", weight_sum );
    PUT_ERR("test 4.) weight deviates from expected value of 1.0!\n");
  }

  yac_correct_weights ( nSourceCells, weight );

  weight_sum = 0.0;
  for ( size_t n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 4.) weight %lf\n", weight_sum );
    PUT_ERR("test 4.) weight deviates from expected value of 1.0!\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 5.) ----------- \n");
#endif

  nSourceCells = 3;

  TargetCell =
    generate_cell_deg((double[]){25.0, 30.0, 30.0, 25.0},
                      (double[]){75.0, 75.0, 80.0, 80.0}, gc_edges, 4);
  SourceCell[0] =
    generate_cell_deg((double[]){30.0, 35.0, 35.0, 30.0},
                      (double[]){75.0, 75.0, 80.0, 80.0}, gc_edges, 4);
  SourceCell[1] =
    generate_cell_deg((double[]){25.0, 30.0, 30.0, 25.0},
                      (double[]){75.0, 75.0, 80.0, 80.0}, gc_edges, 4);
  SourceCell[2] =
    generate_cell_deg((double[]){20.0, 25.0, 25.0, 20.0},
                      (double[]){75.0, 75.0, 80.0, 80.0}, gc_edges, 4);

  yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

  tgt_area = yac_huiliers_area(TargetCell);
  for (size_t n = 0; n < nSourceCells; ++n)
    weight[n] = area[n] / tgt_area;

  for (size_t n = 0; n < nSourceCells; n++) yac_free_grid_cell(SourceCell + n);
  yac_free_grid_cell(&TargetCell);

  weight_sum = 0.0;
  for ( size_t n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 5.) weight %lf\n", weight_sum );
    PUT_ERR("test 5.) weight deviates from expected value of 1.0!\n");
  }

  yac_correct_weights ( nSourceCells, weight );

  weight_sum = 0.0;
  for ( size_t n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 5.) weight %lf\n", weight_sum );
    PUT_ERR("test 5.) weight deviates from expected value of 1.0!\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 6.) ----------- \n");
#endif

  nSourceCells = 1;

  TargetCell =
    generate_cell_rad((double[]){0.098707396293595859,
                                 0.117826295355522680,
                                 0.098174496441371703},
                      (double[]){0.333033372840672190,
                                 0.345484775276814930,
                                 0.353335609258583650}, gc_edges, 3);

  SourceCell[0] =
    generate_cell_rad((double[]){0.073631077818510776,
                                 0.098174770424681035,
                                 0.098174770424681035,
                                 0.073631077818510776},
                      (double[]){0.343611696486383620,
                                 0.343611696486383620,
                                 0.368155389092553910,
                                 0.368155389092553910}, gc_edges, 4);

  yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

  tgt_area = yac_huiliers_area(TargetCell);
  for (size_t n = 0; n < nSourceCells; ++n)
    weight[n] = area[n] / tgt_area;

  for (size_t n = 0; n < nSourceCells; ++n) yac_free_grid_cell(SourceCell + n);
  yac_free_grid_cell(&TargetCell);

  if ( weight[0] < 0 ) {
    printf ("test 6.) weight is smaller than zero (%.2e)\n", weight[0] );
    PUT_ERR("test 6.) weight deviates from expected value > 0.0!\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 7.) ----------- \n");
#endif
  {
    nSourceCells = 4;

    TargetCell =
      generate_cell_deg((double[]){-0.5,  0.5, 0.5, -0.5},
                        (double[]){-0.5, -0.5, 0.5,  0.5}, latlon_edges, 4);
    SourceCell[0] =
      generate_cell_deg((double[]){-1,  0, 0, -1},
                        (double[]){-1, -1, 0,  0}, latlon_edges, 4);
    SourceCell[1] =
      generate_cell_deg((double[]){ 0,  1, 1, 0},
                        (double[]){-1, -1, 0, 0}, latlon_edges, 4);
    SourceCell[2] =
      generate_cell_deg((double[]){-1, 0, 0, -1},
                        (double[]){ 0, 0, 1,  1}, latlon_edges, 4);
    SourceCell[3] =
      generate_cell_deg((double[]){0, 1, 1, 0},
                        (double[]){0, 0, 1, 1}, latlon_edges, 4);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    double ref_tgt_area = yac_pole_area(TargetCell);
    tgt_area = 0.0;
    for (size_t n = 0; n < nSourceCells; ++n)
      tgt_area += area[n];

    for (size_t n = 0; n < nSourceCells; ++n) yac_free_grid_cell(SourceCell + n);
    yac_free_grid_cell(&TargetCell);

    if ( compare_area(ref_tgt_area, tgt_area) )
      PUT_ERR("test 7.) error in partial area computation\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 8.) ----------- \n");
#endif
  {
    nSourceCells = 1;

    TargetCell =
      generate_cell_deg((double[]){-0.5,  0.5, 0.0},
                        (double[]){-0.5, -0.5, 0.5}, gc_edges, 3);
    SourceCell[0] =
      generate_cell_deg((double[]){-1,  1, 1, -1},
                        (double[]){-1, -1, 1,  1}, latlon_edges, 4);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    double ref_tgt_area = yac_pole_area(TargetCell);
    tgt_area = 0.0;
    for (size_t n = 0; n < nSourceCells; ++n)
      tgt_area += area[n];

    for (size_t n = 0; n < nSourceCells; ++n) yac_free_grid_cell(SourceCell + n);
    yac_free_grid_cell(&TargetCell);

    if ( compare_area(ref_tgt_area, tgt_area) )
      PUT_ERR("test 7.) error in partial area computation\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 9.) ----------- \n");
#endif
  {
    nSourceCells = 1;

    TargetCell =
      generate_cell_deg((double[]){-1,  1, 1, -1},
                        (double[]){-1, -1, 1,  1}, latlon_edges, 4);
    SourceCell[0] =
      generate_cell_deg((double[]){-0.5,  0.5, 0.0},
                        (double[]){-0.5, -0.5, 0.5}, gc_edges, 3);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    double ref_area = yac_pole_area(SourceCell[0]);
    double area_sum = 0.0;
    for (size_t n = 0; n < nSourceCells; ++n)
      area_sum += area[n];

  for (size_t n = 0; n < nSourceCells; ++n) yac_free_grid_cell(SourceCell + n);
  yac_free_grid_cell(&TargetCell);

    if ( compare_area(ref_area, area_sum) )
      PUT_ERR("test 7.) error in partial area computation\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 10.) ----------- \n");
#endif
  {
    nSourceCells = 4;

    TargetCell =
      generate_cell_rad((double[]){2.5034566458293663,
                                   2.5280003384355365,
                                   2.5280003384355365,
                                   2.5034566458293663},
                        (double[]){0.5890486225480862,
                                   0.5890486225480862,
                                   0.6135923151542565,
                                   0.6135923151542565},
                        latlon_edges, 4);
    SourceCell[0] =
      generate_cell_rad((double[]){2.5132750564642796,
                                   2.5132751216401110,
                                   2.5530451220055963},
                        (double[]){0.6227096831849960,
                                   0.5888365165503746,
                                   0.6044728875052012}, gc_edges, 3);
    SourceCell[1] =
      generate_cell_rad((double[]){2.5530451220055963,
                                   2.5132751216401110,
                                   2.5514531161521616},
                        (double[]){0.6044728875052012,
                                   0.5888365165503746,
                                   0.5706926293017953}, gc_edges, 3);
    SourceCell[2] =
      generate_cell_rad((double[]){2.4735050741253155,
                                   2.5132751216401110,
                                   2.5132750564642796},
                        (double[]){0.6044725603601205,
                                   0.5888365165503746,
                                   0.6227096831849960}, gc_edges, 3);
    SourceCell[3] =
      generate_cell_rad((double[]){2.4750971948235070,
                                   2.5132751216401110,
                                   2.4735050741253155},
                        (double[]){0.5706923461529829,
                                   0.5888365165503746,
                                   0.6044725603601205}, gc_edges, 3);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    double ref_area = yac_pole_area(TargetCell);
    double area_sum = 0.0;
    for (size_t n = 0; n < nSourceCells; ++n)
      area_sum += area[n];

    for (size_t n = 0; n < nSourceCells; ++n) yac_free_grid_cell(SourceCell + n);
    yac_free_grid_cell(&TargetCell);

    if ( compare_area(ref_area, area_sum) )
      PUT_ERR("test 7.) error in partial area computation\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 11.) ----------- \n");
#endif
  {
    nSourceCells = 3;

    TargetCell =
      generate_cell_rad((double[]){0.6626797003665970,
                                   0.6872233929727672,
                                   0.6872233929727672,
                                   0.6626797003665970},
                        (double[]){1.2271846303085130,
                                   1.2271846303085130,
                                   1.2517283229146832,
                                   1.2517283229146832}, latlon_edges, 4);

    SourceCell[0] =
      generate_cell_rad((double[]){0.6910166804158336,
                                   0.6283648515564006,
                                   0.7428780447168533},
                        (double[]){1.2515918327919406,
                                   1.2207207913058662,
                                   1.2190404189581385}, gc_edges, 3);
    SourceCell[1] =
      generate_cell_rad((double[]){0.5657152198049494,
                                   0.6283648515564006,
                                   0.6910166804158336},
                        (double[]){1.2515932566235379,
                                   1.2207207913058662,
                                   1.2515918327919406}, gc_edges, 3);
    SourceCell[2] =
      generate_cell_rad((double[]){0.6283674880262698,
                                   0.5657152198049494,
                                   0.6910166804158336},
                        (double[]){1.2831450892755811,
                                   1.2515932566235379,
                                   1.2515918327919406}, gc_edges, 3);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    double ref_area = yac_pole_area(TargetCell);
    double area_sum = 0.0;
    for (size_t n = 0; n < nSourceCells; ++n)
      area_sum += area[n];

    for (size_t n = 0; n < nSourceCells; ++n) yac_free_grid_cell(SourceCell + n);
    yac_free_grid_cell(&TargetCell);

    if ( compare_area(ref_area, area_sum) )
      PUT_ERR("test 7.) error in partial area computation\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 12.) ----------- \n");
#endif
  {
    nSourceCells = 7;

    TargetCell =
      generate_cell_rad((double[]){-2.1798479726998332,
                                   -2.1991148570983681,
                                   -2.2190472361409284},
                        (double[]){-0.0311624518327736,
                                    0.0000000043633749,
                                   -0.0338654325037920}, gc_edges, 3);
    SourceCell[0] =
      generate_cell_rad((double[]){ 4.0987966652304335,
                                    4.1233403578366037,
                                    4.1233403578366037,
                                    4.0987966652304335},
                        (double[]){-0.0490873852123405,
                                   -0.0490873852123405,
                                   -0.0245436926061703,
                                   -0.0245436926061703}, latlon_edges, 4);
    SourceCell[1] =
      generate_cell_rad((double[]){ 4.0742529726242633,
                                    4.0987966652304335,
                                    4.0987966652304335,
                                    4.0742529726242633},
                        (double[]){-0.0490873852123405,
                                   -0.0490873852123405,
                                   -0.0245436926061703,
                                   -0.0245436926061703}, latlon_edges, 4);
    SourceCell[2] =
      generate_cell_rad((double[]){ 4.0742529726242633,
                                    4.0987966652304335,
                                    4.0987966652304335,
                                    4.0742529726242633},
                        (double[]){ 0.0000000000000000,
                                    0.0000000000000000,
                                    0.0245436926061703,
                                    0.0245436926061703}, latlon_edges, 4);
    SourceCell[3] =
      generate_cell_rad((double[]){ 4.0987966652304335,
                                    4.1233403578366037,
                                    4.1233403578366037,
                                    4.0987966652304335},
                        (double[]){-0.0245436926061703,
                                   -0.0245436926061703,
                                    0.0000000000000000,
                                    0.0000000000000000}, latlon_edges, 4);
    SourceCell[4] =
      generate_cell_rad((double[]){ 4.0742529726242633,
                                    4.0987966652304335,
                                    4.0987966652304335,
                                    4.0742529726242633},
                        (double[]){-0.0245436926061703,
                                   -0.0245436926061703,
                                    0.0000000000000000,
                                    0.0000000000000000}, latlon_edges, 4);
    SourceCell[5] =
      generate_cell_rad((double[]){ 4.0497092800180932,
                                    4.0742529726242633,
                                    4.0742529726242633,
                                    4.0497092800180932},
                        (double[]){-0.0245436926061703,
                                   -0.0245436926061703,
                                    0.0000000000000000,
                                    0.0000000000000000}, latlon_edges, 4);
    SourceCell[6] =
      generate_cell_rad((double[]){ 4.0497092800180932,
                                    4.0742529726242633,
                                    4.0742529726242633,
                                    4.0497092800180932},
                        (double[]){-0.0490873852123405,
                                   -0.0490873852123405,
                                   -0.0245436926061703,
                                   -0.0245436926061703}, latlon_edges, 4);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    double ref_area = yac_pole_area(TargetCell);
    double area_sum = 0.0;
    for (size_t n = 0; n < nSourceCells; ++n)
      area_sum += area[n];

    for (size_t n = 0; n < nSourceCells; ++n) yac_free_grid_cell(SourceCell + n);
    yac_free_grid_cell(&TargetCell);

    if ( compare_area(ref_area, area_sum) )
      PUT_ERR("test 12.) error in partial area computation\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 13.) ----------- \n");
#endif
  {
    nSourceCells = 4;

    TargetCell =
      generate_cell_rad((double[]){ 3.4361169648638361,
                                    3.4606606574700067,
                                    3.4606606574700067,
                                    3.4361169648638361},
                        (double[]){-0.0245436926061703,
                                   -0.0245436926061703,
                                    0.0000000000000000,
                                    0.0000000000000000}, latlon_edges, 4);

    SourceCell[0] =
      generate_cell_rad((double[]){-2.8075010104343816,
                                   -2.8274333887952685,
                                   -2.8467002736213392},
                        (double[]){-0.0338654341705874,
                                    0.0000000011622542,
                                   -0.0311624540855508}, gc_edges, 3);
    SourceCell[1] =
      generate_cell_rad((double[]){-2.8081665040027728,
                                   -2.8274333887952685,
                                   -2.7882305020025413},
                        (double[]){ 0.0311624566805337,
                                    0.0000000011622542,
                                   -0.0026744729717723}, gc_edges, 3);
    SourceCell[2] =
      generate_cell_rad((double[]){-2.7882305020025413,
                                   -2.8274333887952685,
                                   -2.8075010104343816},
                        (double[]){-0.0026744729717723,
                                    0.0000000011622542,
                                   -0.0338654341705874}, gc_edges, 3);
    SourceCell[3] =
      generate_cell_rad((double[]){-2.8467002736213392,
                                   -2.8274333887952685,
                                   -2.8666362753717674},
                        (double[]){-0.0311624540855508,
                                    0.0000000011622542,
                                    0.0026744751289656}, gc_edges, 3);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    double ref_area = yac_pole_area(TargetCell);
    double area_sum = 0.0;
    for (size_t n = 0; n < nSourceCells; ++n)
      area_sum += area[n];

    for (size_t n = 0; n < nSourceCells; ++n) yac_free_grid_cell(SourceCell + n);
    yac_free_grid_cell(&TargetCell);

    if ( compare_area(ref_area, area_sum) )
      PUT_ERR("test 13.) error in partial area computation\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 14.) ----------- \n");
#endif
  {
    nSourceCells = 3;

    TargetCell =
      generate_cell_rad((double[]){ 3.1170489609836229,
                                    3.1415926535897931,
                                    3.1415926535897931,
                                    3.1170489609836229},
                        (double[]){-1.0062913968529805,
                                   -1.0062913968529805,
                                   -0.9817477042468103,
                                   -0.9817477042468103}, latlon_edges, 4);
    SourceCell[0] =
      generate_cell_rad((double[]){ 3.1415926521557882,
                                    3.1415926520689506,
                                   -3.0774466652769172},
                        (double[]){-0.9805860393425995,
                                   -1.0172219678978514,
                                   -0.9979427097227050}, gc_edges, 3);
    SourceCell[1] =
      generate_cell_rad((double[]){ 3.0774466624490269,
                                    3.1415926520689506,
                                    3.1415926521557882},
                        (double[]){-0.9979427096430752,
                                   -1.0172219678978514,
                                   -0.9805860393425995}, gc_edges, 3);
    SourceCell[2] =
      generate_cell_rad((double[]){ 3.0808931330453300,
                                    3.0774466624490269,
                                    3.1415926521557882},
                        (double[]){-0.9613498550843829,
                                   -0.9979427096430752,
                                   -0.9805860393425995}, gc_edges, 3);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    double ref_area = yac_pole_area(TargetCell);
    double area_sum = 0.0;
    for (size_t n = 0; n < nSourceCells; ++n)
      area_sum += area[n];

    for (size_t n = 0; n < nSourceCells; ++n) yac_free_grid_cell(SourceCell + n);
    yac_free_grid_cell(&TargetCell);

    if ( compare_area(ref_area, area_sum) )
      PUT_ERR("test 14.) error in partial area computation\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 15.) ----------- \n");
#endif
  {
    nSourceCells = 198;

    TargetCell =
      generate_cell_deg((double[]){22.8313,
                                   49.1687,
                                   36.0000},
                        (double[]){89.7418,
                                   89.7418,
                                   89.8335}, gc_edges, 3);

    double dx = ((49.2 - 22.8) /((double)nSourceCells));

    for (size_t i = 0; i < nSourceCells; ++i) {
      SourceCell[i] =
        generate_cell_deg((double[]){22.8000 + dx*((double)(i+0)),
                                     22.8000 + dx*((double)(i+1)),
                                     22.8000 + dx*((double)(i+1)),
                                     22.8000 + dx*((double)(i+0))},
                          (double[]){89.7333,
                                     89.7333,
                                     89.8667,
                                     89.8667}, latlon_edges, 4);
    }

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    double ref_area = yac_pole_area(TargetCell);
    double area_sum = 0.0;
    for (size_t n = 0; n < nSourceCells; ++n)
      area_sum += area[n];

    for (size_t n = 0; n < nSourceCells; ++n) yac_free_grid_cell(SourceCell + n);
    yac_free_grid_cell(&TargetCell);

    if ( compare_area(ref_area, area_sum) )
      PUT_ERR("test 15.) error in partial area computation\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 16.) ----------- \n");
#endif
  {
    TargetCell =
      generate_cell_rad((double[]){ 5.0069132916587327,
                                    5.0130492148102759,
                                    5.0130492148102759,
                                    5.0069132916587327},
                        (double[]){-1.2946797849754812,
                                   -1.2946797849754812,
                                   -1.2885438618239387,
                                   -1.2885438618239387}, latlon_edges, 4);
    SourceCell[0] =
      generate_cell_rad((double[]){ 4.9999997254688875,
                                    4.9999997254688875,
                                    5.0078539201557488,
                                    5.0078539201557488},
                        (double[]){-1.2878426515089207,
                                   -1.2946756570757916,
                                   -1.2946797849754812,
                                   -1.2878469125666649}, gc_edges, 4);
    SourceCell[1] =
      generate_cell_rad((double[]){ 5.0078539201557488,
                                    5.0078539201557488,
                                    5.0157075822103927,
                                    5.0157075822103927},
                        (double[]){-1.2946797849754812,
                                   -1.3015127905423520,
                                   -1.3015170516000960,
                                   -1.2946840460332254}, gc_edges, 4);
    SourceCell[2] =
      generate_cell_rad((double[]){ 5.0078539201557488,
                                    5.0078539201557488,
                                    5.0157075822103927,
                                    5.0157075822103927},
                        (double[]){-1.2878469125666649,
                                   -1.2946797849754812,
                                   -1.2946840460332254,
                                   -1.2878513067824635}, gc_edges, 4);
    SourceCell[3] =
      generate_cell_rad((double[]){ 4.9999997254688875,
                                    4.9999997254688875,
                                    5.0078539201557488,
                                    5.0078539201557488},
                        (double[]){-1.2946756570757916,
                                   -1.3015085294846078,
                                   -1.3015127905423520,
                                   -1.2946797849754812}, gc_edges, 4);

    nSourceCells = 4;

    double src_area[nSourceCells];

    yac_compute_overlap_areas(
      nSourceCells, SourceCell, TargetCell, src_area);

    double ref_area = yac_pole_area(TargetCell);
    double area_sum = 0.0;
    for (unsigned n = 0; n < nSourceCells; ++n)
      area_sum += src_area[n];

    for (size_t n = 0; n < nSourceCells; ++n) yac_free_grid_cell(SourceCell + n);
    yac_free_grid_cell(&TargetCell);

    if ( compare_area(ref_area, area_sum) )
      PUT_ERR("test 16.) error in partial area computation\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 17.) ----------- \n");
#endif
  {
    TargetCell =
      generate_cell_rad((double[]){-2.8460897445189417,
                                   -2.8659852121839959,
                                   -2.8853055611946927},
                        (double[]){-0.0961602003314924,
                                   -0.0622820530745658,
                                   -0.0933150814318732}, gc_edges, 3);
    SourceCell[0] =
      generate_cell_rad((double[]){-2.8655528147976832,
                                   -2.9038195735214831,
                                   -2.8254194450162888},
                        (double[]){-0.0615856127966505,
                                   -0.1228534567110898,
                                   -0.1292806814915037}, gc_edges, 3);
    SourceCell[1] =
      generate_cell_rad((double[]){-2.9440830021466091,
                                   -2.9038195735214831,
                                   -2.8655528147976832},
                        (double[]){-0.0558979744839555,
                                   -0.1228534567110898,
                                   -0.0615856127966505}, gc_edges, 3);

    nSourceCells = 2;

    double src_area[nSourceCells];

    yac_compute_overlap_areas(nSourceCells, SourceCell, TargetCell, src_area );

    double ref_area = yac_pole_area(TargetCell);
    double area_sum = 0.0;
    for (unsigned n = 0; n < nSourceCells; ++n)
      area_sum += src_area[n];

    for (size_t n = 0; n < nSourceCells; ++n) yac_free_grid_cell(SourceCell + n);
    yac_free_grid_cell(&TargetCell);

    if ( compare_area(ref_area, area_sum) )
      PUT_ERR("test 17.) error in partial area computation\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 18.) ----------- \n");
#endif
  {
    TargetCell =
      generate_cell_rad((double[]){3.0127596408133406,
                                   2.9702182441873108,
                                   2.9943996418061438},
                        (double[]){0.5498596133493898,
                                   0.5469967117356207,
                                   0.5140000458840724}, gc_edges, 3);
    SourceCell[0] =
      generate_cell_rad((double[]){2.9198084624304621,
                                   2.9720969737232750,
                                   3.0078074923914948},
                        (double[]){0.6123031304013200,
                                   0.5471403770103002,
                                   0.6192726969308253}, gc_edges, 3);
    SourceCell[1] =
      generate_cell_rad((double[]){2.8889398582596519,
                                   2.9720969737232750,
                                   2.9198084624304621},
                        (double[]){0.5392518296263771,
                                   0.5471403770103002,
                                   0.6123031304013200}, gc_edges, 3);
    SourceCell[2] =
      generate_cell_rad((double[]){3.0187962726049782,
                                   2.9720969737232750,
                                   2.9378118571213374},
                        (double[]){0.4805223860987507,
                                   0.5471403770103002,
                                   0.4745836829538209}, gc_edges, 3);
    SourceCell[3] =
      generate_cell_rad((double[]){2.9378118571213374,
                                   2.9720969737232750,
                                   2.8889398582596519},
                        (double[]){0.4745836829538209,
                                   0.5471403770103002,
                                   0.5392518296263771}, gc_edges, 3);
    SourceCell[4] =
      generate_cell_rad((double[]){3.0565195473916882,
                                   2.9720969737232750,
                                   3.0187962726049782},
                        (double[]){0.5519553785110132,
                                   0.5471403770103002,
                                   0.4805223860987507}, gc_edges, 3);
    SourceCell[5] =
      generate_cell_rad((double[]){3.0078074923914948,
                                   2.9720969737232750,
                                   3.0565195473916882},
                        (double[]){0.6192726969308253,
                                   0.5471403770103002,
                                   0.5519553785110132}, gc_edges, 3);

    nSourceCells = 6;

    double src_area[nSourceCells];

    yac_compute_overlap_areas(nSourceCells, SourceCell, TargetCell, src_area );

    double ref_area = yac_pole_area(TargetCell);
    double area_sum = 0.0;
    for (unsigned n = 0; n < nSourceCells; ++n)
      area_sum += src_area[n];

    for (size_t n = 0; n < nSourceCells; ++n) yac_free_grid_cell(SourceCell + n);
    yac_free_grid_cell(&TargetCell);

    if ( compare_area(ref_area, area_sum) )
      PUT_ERR("test 18.) error in partial area computation\n");
  }

  free(SourceCell);

  return TEST_EXIT_CODE;
}

static int compare_area(double a, double b) {

  double diff = fabs(a - b);
  double tol = MAX(MIN(a,b)*1e-6, YAC_AREA_TOL);

  if (diff < tol) return 0;
  else return (a > b) - (a < b);
}

